@@ -47,7 +47,10 @@ void dump_db_raw(hd_data_t *hd_data);
 void dump_db(hd_data_t *hd_data);
 void do_chroot(hd_data_t *hd_data, char *dir);
 void ask_db(hd_data_t *hd_data, char *query);
+void get_mapping(hd_data_t *hd_data);
+void write_udi(hd_data_t *hd_data, char *udi);
 
+void do_saveconfig(hd_data_t *hd_data, hd_t *hd, FILE *f);
 
 struct {
   unsigned db_idx;
@@ -75,6 +78,10 @@ struct option options[] = {
   { "db", 1, NULL, 310 },
   { "only", 1, NULL, 311 },
   { "listmd", 0, NULL, 312 },
+  { "map", 0, NULL, 313 },
+  { "kernel-version", 1, NULL, 314 },
+  { "write-udi", 1, NULL, 315 },
+  { "hddb-dir", 1, NULL, 316 },
   { "cdrom", 0, NULL, 1000 + hw_cdrom },
   { "floppy", 0, NULL, 1000 + hw_floppy },
   { "disk", 0, NULL, 1000 + hw_disk },
@@ -133,6 +140,7 @@ struct option options[] = {
   { "block", 0, NULL, 1000 + hw_block },
   { "tape", 0, NULL, 1000 + hw_tape },
   { "vbe", 0, NULL, 1000 + hw_vbe },
+  { "bluetooth", 0, NULL, 1000 + hw_bluetooth },
   { "all", 0, NULL, 2000 },
   { "reallyall", 0, NULL, 2001 },
   { "smp", 0, NULL, 2002 },
@@ -251,6 +259,22 @@ int main(int argc, char **argv)
           hd_data->flags.list_md = 1;
           break;
 
+        case 313:
+          get_mapping(hd_data);
+          break;
+
+        case 314:
+          if(*optarg) setenv("LIBHD_KERNELVERSION", optarg, 1);
+          break;
+
+        case 315:
+          write_udi(hd_data, optarg);
+          break;
+
+        case 316:
+          if(*optarg) setenv("LIBHD_HDDB_DIR", optarg, 1);
+          break;
+
         case 400:
           printf("%s\n", hd_version());
 	  break;
@@ -303,6 +327,12 @@ int main(int argc, char **argv)
 #ifndef LIBHD_TINY
       if(showconfig) {
         hd = hd_read_config(hd_data, showconfig);
+        if(hd_data->debug == -1) {
+          fprintf(f ? f : stdout,
+            "============ start debug info ============\n%s=========== end debug info ============\n",
+            hd_data->log
+          );
+        }
         if(hd) {
           hd_dump_entry(hd_data, hd, f ? f : stdout);
           hd = hd_free_hd_list(hd);
@@ -311,22 +341,6 @@ int main(int argc, char **argv)
           fprintf(f ? f : stdout, "No config data: %s\n", showconfig);
         }
       }
-
-      if(saveconfig) {
-        for(hd = hd_data->hd; hd; hd = hd->next) {
-          if(!strcmp(hd->unique_id, saveconfig)) {
-            i = hd_write_config(hd_data, hd);
-            fprintf(f ? f : stdout, "%s: %s\n",
-              saveconfig,
-              i ? "Error writing config data" : "config saved"
-            );
-            break;
-          }
-        }
-        if(!hd) {
-          fprintf(f ? f : stdout, "No such hardware: %s\n", saveconfig);
-        }
-      }
 #endif
 
       if(f) fclose(f);
@@ -496,9 +510,12 @@ void do_hw(hd_data_t *hd_data, FILE *f, hd_hw_item_t hw_item)
       );
     }
 
+    i = hd_data->debug;
+    hd_data->debug = -1;
     for(hd = hd_data->hd; hd; hd = hd->next) {
       hd_dump_entry(hd_data, hd, f);
     }
+    hd_data->debug = i;
 
     fprintf(f,
       "============ end hardware log ============\n"
@@ -609,6 +626,7 @@ void do_hw(hd_data_t *hd_data, FILE *f, hd_hw_item_t hw_item)
       for(hd = hd0; hd; hd = hd->next) {
         hd_dump_entry(hd_data, hd, f ? f : stdout);
       }
+      do_saveconfig(hd_data, hd0, f ? f : stdout);
     }
   }
 
@@ -623,6 +641,7 @@ void do_hw(hd_data_t *hd_data, FILE *f, hd_hw_item_t hw_item)
 void do_hw_multi(hd_data_t *hd_data, FILE *f, hd_hw_item_t *hw_items)
 {
   hd_t *hd, *hd0;
+  int i;
 
   hd0 = hd_list2(hd_data, hw_items, 1);
 
@@ -642,9 +661,12 @@ void do_hw_multi(hd_data_t *hd_data, FILE *f, hd_hw_item_t *hw_items)
       );
     }
 
+    i = hd_data->debug;
+    hd_data->debug = -1;
     for(hd = hd_data->hd; hd; hd = hd->next) {
       hd_dump_entry(hd_data, hd, f);
     }
+    hd_data->debug = i;
 
     fprintf(f,
       "============ end hardware log ============\n"
@@ -660,6 +682,7 @@ void do_hw_multi(hd_data_t *hd_data, FILE *f, hd_hw_item_t *hw_items)
     for(hd = hd0; hd; hd = hd->next) {
       hd_dump_entry(hd_data, hd, f ? f : stdout);
     }
+    do_saveconfig(hd_data, hd0, f ? f : stdout);
   }
 
   hd_free_hd_list(hd0);
@@ -695,7 +718,7 @@ void do_short(hd_data_t *hd_data, hd_t *hd, FILE *f)
 }
 
 
-#if 1
+#if 0
 typedef struct {
   char *vendor, *model, *driver;
 } scanner_t;
@@ -994,7 +1017,7 @@ static char *scanner_info(hd_t *hd)
 
 void do_test(hd_data_t *hd_data)
 {
-#if 1
+#if 0
   hd_t *hd, *hd0;
   hd_res_t *res;
   driver_info_t *di;
@@ -1296,6 +1319,21 @@ void do_test(hd_data_t *hd_data)
   if(hd_data->log) printf("%s", hd_data->log);
 #endif
 
+#endif
+
+#if 0
+  hd_t *hd;
+  driver_info_t *di;
+  str_list_t *sl;
+
+  hd = hd_list(hd_data, hw_sys, 1, NULL);
+
+  for(di = hd->driver_info; di; di = di->next) {
+    if(di->any.type == di_module) {
+      for(sl = di->module.names; sl; sl = sl->next) printf("%s\n", sl->str);
+    }
+  }
+
 #endif
 }
 
@@ -1312,11 +1350,11 @@ void help()
     "  --dump-db n    dump hardware data base, 0: external, 1: internal\n"
     "  --hw_item      probe for hw_item\n"
     "  hw_item is one of:\n"
-    "    cdrom, floppy, disk, network, gfxcard, framebuffer, monitor, camera,\n"
-    "    mouse, joystick, keyboard, chipcard, sound, isdn, modem, storage-ctrl,\n"
-    "    netcard, printer, tv, dvb, scanner, braille, sys, bios, cpu, partition,\n"
-    "    usb-ctrl, usb, pci, isapnp, ide, scsi, bridge, hub, memory, smp, pppoe,\n"
-    "    pcmcia, pcmcia-ctrl, wlan, zip, dsl, all, reallyall\n\n"
+    "    all, bios, block, bluetooth, braille, bridge, camera, cdrom, chipcard, cpu,\n"
+    "    disk, dsl, dvb, floppy, framebuffer, gfxcard, hub, ide, isapnp, isdn,\n"
+    "    joystick, keyboard, memory, modem, monitor, mouse, netcard, network,\n"
+    "    partition, pci, pcmcia, pcmcia-ctrl, pppoe, printer, scanner, scsi, smp,\n"
+    "    sound, storage-ctrl, sys, tape, tv, usb, usb-ctrl, vbe, wlan, zip\n\n"
     "  Note: debug info is shown only in the log file. (If you specify a\n"
     "  log file the debug level is implicitly set to a reasonable value.)\n"
   );
@@ -1899,6 +1937,13 @@ void ask_db(hd_data_t *hd_data, char *query)
     if(!strcmp(sl->str, "special")) { tag = TAG_SPECIAL; continue; }
     if(!strcmp(sl->str, "pcmcia")) { tag = TAG_PCMCIA; continue; }
 
+    if(sscanf(sl->str, "class=%i%n", &u, &cnt) >= 1 && !sl->str[cnt]) {
+      hd->base_class.id = u >> 16;
+      hd->sub_class.id = (u >> 8) & 0xff;
+      hd->prog_if.id = u & 0xff;
+      continue;
+    }
+
     if(sscanf(sl->str, "vendor=%i%n", &u, &cnt) >= 1 && !sl->str[cnt]) {
       hd->vendor.id = MAKE_ID(tag, u);
       continue;
@@ -1958,3 +2003,155 @@ void ask_db(hd_data_t *hd_data, char *query)
   }
 }
 
+
+int is_same_block_dev(hd_t *hd1, hd_t *hd2)
+{
+  if(!hd1 || !hd2 || hd1 == hd2) return 0;
+
+  if(
+    hd1->base_class.id != hd2->base_class.id ||
+    hd1->sub_class.id != hd2->sub_class.id
+  ) return 0;
+
+  if(
+    !hd1->model ||
+    !hd2->model ||
+    strcmp(hd1->model, hd2->model)
+  ) return 0;
+
+  if(hd1->revision.name || hd2->revision.name) {
+    if(
+      !hd1->revision.name ||
+      !hd2->revision.name ||
+      strcmp(hd1->revision.name, hd2->revision.name)
+    ) return 0;
+  }
+
+  if(hd1->serial || hd2->serial) {
+    if(
+      !hd1->serial ||
+      !hd2->serial ||
+      strcmp(hd1->serial, hd2->serial)
+    ) return 0;
+  }
+
+  return 1;
+}
+
+
+hd_t *get_same_block_dev(hd_t *hd_list, hd_t *hd, hd_status_value_t status)
+{
+  for(; hd_list; hd_list = hd_list->next) {
+    if(hd_list->status.available != status) continue;
+    if(is_same_block_dev(hd_list, hd)) return hd_list;
+  }
+
+  return NULL;
+}
+
+
+void get_mapping(hd_data_t *hd_data)
+{
+  hd_t *hd_manual, *hd, *hd2;
+  struct {
+    hd_t *hd;
+    unsigned unknown:1;
+  } map[256] = { };
+  unsigned maps = 0, u;
+  int broken, first;
+  hd_hw_item_t hw_items[] = { hw_disk, hw_cdrom, 0 };
+
+  hd_data->progress = NULL;
+
+  hd_data->flags.list_all = 1;
+
+  hd_manual = hd_list2(hd_data, hw_items, 1);
+  for(hd = hd_manual; hd && maps < sizeof map / sizeof *map; hd = hd->next) {
+    if(!hd->unix_dev_name) continue;
+
+    if(hd->status.available == status_yes) {
+      /* check if we already have an active device with the same name */
+      for(broken = u = 0; u < maps; u++) {
+        if(!strcmp(map[u].hd->unix_dev_name, hd->unix_dev_name)) {
+          map[u].unknown = 1;
+          broken = 1;
+        }
+      }
+      if(broken) continue;
+
+      /* ensure we really can tell different devices apart */
+      if(get_same_block_dev(hd_manual, hd, status_yes)) {
+        map[maps].hd = hd;
+        map[maps].unknown = 1;
+      }
+      else {
+        map[maps].hd = hd;
+      }
+      maps++;
+    }
+  }
+
+  /* ok, we have a list of all new devs */
+
+  for(u = 0; u < maps; u++) {
+    if(map[u].unknown) {
+      printf("%s\n", map[u].hd->unix_dev_name);
+    }
+    else {
+      first = 1;
+      for(hd2 = hd_manual; (hd2 = get_same_block_dev(hd2, map[u].hd, status_no)); hd2 = hd2->next) {
+        if(hd2->unix_dev_name && strcmp(map[u].hd->unix_dev_name, hd2->unix_dev_name)) {
+          printf("%s\t%s", first ? map[u].hd->unix_dev_name : "", hd2->unix_dev_name);
+          first = 0;
+        }
+      }
+      if(!first) printf("\n");
+    }
+
+  }
+}
+
+
+void write_udi(hd_data_t *hd_data, char *udi)
+{
+  hal_prop_t prop = {};
+  int i;
+
+  prop.type = p_string;
+  prop.key = "foo.bar";
+  prop.val.str = "test XXX";
+
+  i = hd_write_properties(udi, &prop);
+
+  fprintf(stderr, "write = %d\n", i);
+}
+
+
+
+void do_saveconfig(hd_data_t *hd_data, hd_t *hd, FILE *f)
+{
+#ifndef LIBHD_TINY
+  int i;
+
+  if(!saveconfig) return;
+
+  fprintf(f, "\nSave Configuration:\n");
+  for(; hd; hd = hd->next) {
+    if(
+      !strcmp(saveconfig, "all") ||
+      (hd->udi && !strcmp(hd->udi, saveconfig)) ||
+      (hd->unique_id && !strcmp(hd->unique_id, saveconfig)) ||
+      (hd->unix_dev_name && !strcmp(hd->unix_dev_name, saveconfig)) ||
+      (hd->unix_dev_name2 && !strcmp(hd->unix_dev_name2, saveconfig)) ||
+      search_str_list(hd->unix_dev_names, saveconfig)
+    ) {
+      i = hd_write_config(hd_data, hd);
+      fprintf(f, "  %s: %s\n",
+        hd->udi ?: hd->unique_id ?: saveconfig,
+        i ? "failed" : "ok"
+      );
+    }
+  }
+#endif
+}
+
