@@ -23,39 +23,51 @@
  * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  */
 
-#if defined(__i386__) || defined (__x86_64__)
+#if defined(__i386__) || defined (__x86_64__) || defined (__ia64__)
+
+#include "ibm-notebooks.h"
+static int tp_lookup(char *key, unsigned *width, unsigned *height, unsigned *xsize, unsigned *ysize);
 
-#ifndef LIBHD_TINY
 static struct {
-  int width;
-  int height;
+  unsigned short xsize;	/* mm */
+  unsigned short ysize;	/* mm */
+  unsigned short width;
+  unsigned short height;
   char *vendor;
   char *name;
   char *version;
-} panel_data[] = {
-  {  800,  600, "Fujitsu Siemens", "LiteLine", "LF6" },
-  { 1024,  768, "ASUSTEK", "L2000D", NULL },
-  { 1024,  768, "ASUSTeK Computer Inc.", "L8400C series Notebook PC", NULL },
-  { 1024,  768, "Acer", "TravelMate 720", NULL },
-  { 1024,  768, "COMPAL", "N30T5", NULL },
-  { 1024,  768, "Dell Computer Corporation", "Inspiron 5000", NULL },
-  { 1024,  768, "Dell Computer Corporation", "Latitude C400", NULL },
-  { 1024,  768, "Dell Computer Corporation", "Latitude CPt C400GT", NULL },
-  { 1024,  768, "Hewlett-Packard", "HP OmniBook PC", "HP OmniBook 4150 B" },
-  { 1024,  768, "IBM", "2373G1U", NULL },
-  { 1024,  768, "IBM", "26284UG", NULL },
-  { 1024,  768, "IBM", "26446AG", NULL },
-  { 1024,  768, "IBM", "26455BG", NULL },
-  { 1024,  768, "IBM", "264721G", NULL },
-  { 1024,  768, "KDST", "KDS6KSUMO", NULL  },
-  { 1024,  768, "Sony Corporation", "PCG-F370(UC)", NULL },
-  { 1024,  768, "Sony Corporation", "PCG-N505SN", NULL },
-  { 1024,  768, "TOSHIBA", "S2400-103", NULL },
-  { 1400, 1050, "Dell Computer Corporation", "Inspiron 8000", NULL },
-  { 1600, 1200, "Dell Computer Corporation", "Inspiron 8200", NULL },
-  { 1600, 1200, "Dell Computer Corporation", "Latitude C840", NULL }
+} __attribute((packed)) panel_data[] = {
+  {   0,   0, 1400, 1050, "IBM", "73geu99", NULL },
+  {   0,   0,  800,  600, "Fujitsu Siemens", "LiteLine", "LF6" },
+  {   0,   0, 1024,  768, "ASUSTEK", "L2000D", NULL },
+  {   0,   0, 1024,  768, "ASUSTeK Computer Inc.", "L8400C series Notebook PC", NULL },
+  {   0,   0, 1024,  768, "ASUSTeK Computer Inc.", "S5N", NULL },
+  {   0,   0, 1024,  768, "Acer", "TravelMate 720", NULL },
+  {   0,   0, 1024,  768, "COMPAL", "N30T5", NULL },
+  {   0,   0, 1024,  768, "Dell Computer Corporation", "Inspiron 5000", NULL },
+  {   0,   0, 1024,  768, "Dell Computer Corporation", "Latitude C400", NULL },
+  {   0,   0, 1024,  768, "Dell Computer Corporation", "Latitude C600", NULL },
+  {   0,   0, 1024,  768, "Dell Computer Corporation", "Latitude CPt C400GT", NULL },
+  {   0,   0, 1024,  768, "Dell Computer Corporation", "Latitude CPx J650GT", NULL },
+  {   0,   0, 1024,  768, "Hewlett-Packard", "HP OmniBook PC", "HP OmniBook 4150 B" },
+  {   0,   0, 1280,  800, "Hewlett-Packard", "hp compaq nx9105 (DU367T#ABD)", "03" },
+  { 330, 210, 1280,  800, "Hewlett-Packard", "hp compaq nx9105 (DU367T#ABD)", "F.21" },
+  {   0,   0, 1280,  800, "Hewlett-Packard", "Pavilion zv5000 (PA456EA#ABD)", "F.11" },
+  {   0,   0, 1024,  768, "KDST", "KDS6KSUMO", NULL  },
+  {   0,   0, 1024,  768, "Sony Corporation", "PCG-F370(UC)", NULL },
+  {   0,   0, 1024,  768, "Sony Corporation", "PCG-N505SN", NULL },
+  {   0,   0, 1024,  768, "TOSHIBA", "S2400-103", NULL },
+  {   0,   0, 1280,  800, "Acer", "Aspire 1520", NULL },
+  {   0,   0, 1440,  900, "FUJITSU SIEMENS", "Amilo M3438 Series", NULL },
+  {   0,   0, 1400, 1050, "Acer", "TravelMate 660", NULL },
+  {   0,   0, 1400, 1050, "Dell Computer Corporation", "Inspiron 8000", NULL },
+  { 286, 214, 1400, 1050, "Dell Computer Corporation", "Latitude D600", NULL },
+  {   0,   0, 1400, 1050, "TOSHIBA", "TECRA 9100", NULL },
+  {   0,   0, 1600, 1200, "Dell Computer Corporation", "Inspiron 8200", NULL },
+  {   0,   0, 1600, 1200, "Dell Computer Corporation", "Latitude C840", NULL },
+  { 190, 110, 1024,  600, "FUJITSU SIEMENS", "LIFEBOOK P1510", NULL },
+  { 285, 215, 1024,  768, "FUJITSU SIEMENS", "LIFEBOOK S7020", NULL }
 };
-#endif
 
 #define BIOS_TEST
 
@@ -63,24 +75,26 @@ typedef struct {
   unsigned eax, ebx, ecx, edx, esi, edi, eip, es, iret, cli;
 } bios32_regs_t;
 
-static void read_memory(memory_range_t *mem);
-#ifndef LIBHD_TINY
+static void read_memory(hd_data_t *hd_data, memory_range_t *mem);
 static void dump_memory(hd_data_t *hd_data, memory_range_t *mem, int sparse, char *label);
 static void get_pnp_support_status(memory_range_t *mem, bios_info_t *bt);
 static void smbios_get_info(hd_data_t *hd_data, memory_range_t *mem, bios_info_t *bt);
 static void get_fsc_info(hd_data_t *hd_data, memory_range_t *mem, bios_info_t *bt);
+#ifndef LIBHD_TINY
 static void add_panel_info(hd_data_t *hd_data, bios_info_t *bt);
+#endif
 static void add_mouse_info(hd_data_t *hd_data, bios_info_t *bt);
 static unsigned char crc(unsigned char *mem, unsigned len);
 static int get_smp_info(hd_data_t *hd_data, memory_range_t *mem, smp_info_t *smp);
 static void parse_mpconfig(hd_data_t *hd_data, memory_range_t *mem, smp_info_t *smp);
 static int get_bios32_info(hd_data_t *hd_data, memory_range_t *mem, bios32_info_t *bios32);
-#endif
 
-int detect_smp(hd_data_t *hd_data)
+int detect_smp_bios(hd_data_t *hd_data)
 {
   bios_info_t *bt;
+  hd_smbios_t *sm;
   hd_t *hd;
+  int cpus;
 
   if(!hd_data->bios_ram.data) return -1;	/* hd_scan_bios() not called */
 
@@ -98,12 +112,25 @@ int detect_smp(hd_data_t *hd_data)
 
   if(!bt) return -1;
 
-//  return bt->smp.ok ? bt->smp.cpus_en ? bt->smp.cpus_en : 1 : 0;
-// Dell Dimension 8100 has a MP table with 0 cpus!
+  cpus = 0;
 
-  return bt->smp.ok ? bt->smp.cpus_en : 0;
+  /* look at smbios data in case there's no mp table */
+  if(hd_data->smbios) {
+    for(sm = hd_data->smbios; sm; sm = sm->next) {
+      if(
+        sm->any.type == sm_processor &&
+        sm->processor.pr_type.id == 3 &&	/* cpu */
+        sm->processor.cpu_status.id == 1	/* enabled */
+      ) {
+        cpus++;
+      }
+    }
+    ADD2LOG("  smp detect: mp %d cpus, smbios %d cpus\n", bt->smp.ok ? bt->smp.cpus_en : 0, cpus);
+  }
 
-  return 0;
+  if(bt->smp.ok && bt->smp.cpus_en) cpus = bt->smp.cpus_en;
+
+  return cpus;
 }
 
 
@@ -111,23 +138,22 @@ void hd_scan_bios(hd_data_t *hd_data)
 {
   hd_t *hd;
   bios_info_t *bt;
-#ifndef LIBHD_TINY
   char *s;
   unsigned char *bios_ram;
   unsigned u, u1;
   memory_range_t mem;
   unsigned smp_ok;
+#ifndef LIBHD_TINY
   vbe_info_t *vbe;
   vbe_mode_info_t *mi;
   hd_res_t *res;
-#endif
-  struct stat sbuf;
   str_list_t *sl;
+#endif
 
   if(!hd_probe_feature(hd_data, pr_bios)) return;
 
   /* we better do nothing on a SGI Altix machine */
-  if(!stat("/proc/sgi_sn", &sbuf)) return;
+  if(hd_is_sgi_altix(hd_data)) return;
 
   hd_data->module = mod_bios;
 
@@ -143,8 +169,6 @@ void hd_scan_bios(hd_data_t *hd_data)
   hd->detail->type = hd_detail_bios;
   hd->detail->bios.data = bt = new_mem(sizeof *bt);
 
-#ifndef LIBHD_TINY
-
   /*
    * first, look for APM support
    */
@@ -200,8 +224,6 @@ void hd_scan_bios(hd_data_t *hd_data)
     free_str_list(sl0);
   }
 
-#endif		/* !defined(LIBHD_TINY) */
-
   /*
    * get the i/o ports for the parallel & serial interfaces from the BIOS
    * memory area starting at 0x40:0
@@ -210,13 +232,11 @@ void hd_scan_bios(hd_data_t *hd_data)
 
   hd_data->bios_ram.start = BIOS_RAM_START;
   hd_data->bios_ram.size = BIOS_RAM_SIZE;
-  read_memory(&hd_data->bios_ram);
+  read_memory(hd_data, &hd_data->bios_ram);
 
   hd_data->bios_rom.start = BIOS_ROM_START;
   hd_data->bios_rom.size = BIOS_ROM_SIZE;
-  read_memory(&hd_data->bios_rom);
-
-#ifndef LIBHD_TINY
+  read_memory(hd_data, &hd_data->bios_rom);
 
   if(hd_data->bios_ram.data) {
     bios_ram = hd_data->bios_ram.data;
@@ -277,7 +297,12 @@ void hd_scan_bios(hd_data_t *hd_data)
 
     bt->low_mem_size = ((bios_ram[0x14] << 8) + bios_ram[0x13]) << 10;
 
-    if(bt->low_mem_size >= 768 || bt->low_mem_size < 500) {
+    if(bt->low_mem_size) {
+      ADD2LOG("  bios: %uk low mem\n", bt->low_mem_size >> 10);
+    }
+
+    /* too unusual */
+    if(bt->low_mem_size >= (768 << 10) || bt->low_mem_size < (384 << 10)) {
       bt->low_mem_size = 0;
     }
 
@@ -287,14 +312,14 @@ void hd_scan_bios(hd_data_t *hd_data)
     if(u) {
       hd_data->bios_ebda.start = u;
       hd_data->bios_ebda.size = 1;	/* just one byte */
-      read_memory(&hd_data->bios_ebda);
+      read_memory(hd_data, &hd_data->bios_ebda);
       if(hd_data->bios_ebda.data) {
         u1 = hd_data->bios_ebda.data[0];
         if(u1 > 0 && u1 <= 64) {	/* be sensible, typically only 1k */
           u1 <<= 10;
           if(u + u1 <= (1 << 20)) {
             hd_data->bios_ebda.size = u1;
-            read_memory(&hd_data->bios_ebda);
+            read_memory(hd_data, &hd_data->bios_ebda);
           }
         }
       }
@@ -317,7 +342,9 @@ void hd_scan_bios(hd_data_t *hd_data)
     get_pnp_support_status(&hd_data->bios_rom, bt);
     smbios_get_info(hd_data, &hd_data->bios_rom, bt);
     get_fsc_info(hd_data, &hd_data->bios_rom, bt);
+#ifndef LIBHD_TINY
     add_panel_info(hd_data, bt);
+#endif
     add_mouse_info(hd_data, bt);
   }
 
@@ -342,7 +369,7 @@ void hd_scan_bios(hd_data_t *hd_data)
     mem.size = 1 << 10;
     mem.start = 639 << 10;
     mem.data = NULL;
-    read_memory(&mem);
+    read_memory(hd_data, &mem);
     if(mem.data) smp_ok = get_smp_info(hd_data, &mem, &bt->smp);
     mem.data = free_mem(mem.data);
   }
@@ -351,7 +378,7 @@ void hd_scan_bios(hd_data_t *hd_data)
     mem.start = bt->smp.mpconfig;
     mem.size = 1 << 16;
     mem.data = NULL;
-    read_memory(&mem);
+    read_memory(hd_data, &mem);
     parse_mpconfig(hd_data, &mem, &bt->smp);
     mem.data = free_mem(mem.data);
   }
@@ -365,7 +392,7 @@ void hd_scan_bios(hd_data_t *hd_data)
       mem.start = bt->smp.mpfp;
       mem.size = 0x10;
       mem.data = NULL;
-      read_memory(&mem);
+      read_memory(hd_data, &mem);
       dump_memory(hd_data, &mem, 0, "MP FP");
       mem.data = free_mem(mem.data);
     }
@@ -374,12 +401,13 @@ void hd_scan_bios(hd_data_t *hd_data)
       mem.start = bt->smp.mpconfig;
       mem.size = bt->smp.mpconfig_size;
       mem.data = NULL;
-      read_memory(&mem);
+      read_memory(hd_data, &mem);
       dump_memory(hd_data, &mem, 0, "MP config table");
       mem.data = free_mem(mem.data);
     }
   }
 
+#ifndef LIBHD_TINY
   if(hd_probe_feature(hd_data, pr_bios_vesa)) {
     PROGRESS(4, 0, "vbe");
 
@@ -465,6 +493,7 @@ void hd_scan_bios(hd_data_t *hd_data)
 
     }
   }
+#endif	/* LIBHD_TINY */
 
   PROGRESS(5, 0, "32");
 
@@ -489,15 +518,11 @@ void hd_scan_bios(hd_data_t *hd_data)
     }
   }
 
-
-#endif		/* !defined(LIBHD_TINY) */
-
 }
 
 
-void read_memory(memory_range_t *mem)
+void read_memory(hd_data_t *hd_data, memory_range_t *mem)
 {
-  int fd;
 #ifdef BIOS_TEST
   char *s = getenv("LIBHD_MEM");
 #endif
@@ -510,19 +535,11 @@ void read_memory(memory_range_t *mem)
 
   if(mem->data) free_mem(mem->data);
   mem->data = new_mem(mem->size);
-  fd = -1;
-  if(
 #ifdef BIOS_TEST
-    (fd = open(s ? s : DEV_MEM, O_RDONLY)) >= 0 &&
+  hd_read_mmap(hd_data, s ?: DEV_MEM, mem->data, mem->start, mem->size);
 #else
-    (fd = open(DEV_MEM, O_RDONLY)) >= 0 &&
+  hd_read_mmap(hd_data, DEV_MEM, mem->data, mem->start, mem->size);
 #endif
-    lseek(fd, mem->start, SEEK_SET) >= 0
-  ) {
-    read(fd, mem->data, mem->size);
-  }
-
-  if(fd >= 0) close(fd);
 
 #ifdef LIBHD_MEMCHECK
   {
@@ -547,14 +564,13 @@ void dump_memory(hd_data_t *hd_data, memory_range_t *mem, int sparse, char *labe
   ADD2LOG("----- %s 0x%05x - 0x%05x -----\n", label, mem->start, mem->start + mem->size - 1);
   for(u = 0; u < mem->size; u += step) {
     ADD2LOG("  %03x  ", u + mem->start);
-    hexdump(&hd_data->log, 1, 0x10, mem->data + u);
+    hexdump(&hd_data->log, 1, mem->size - u > 0x10 ? 0x10 : mem->size - u, mem->data + u);
     ADD2LOG("\n");
   }
   ADD2LOG("----- %s end -----\n", label);
 }
 
 
-#ifndef LIBHD_TINY
 void get_pnp_support_status(memory_range_t *mem, bios_info_t *bt)
 {
   int i;
@@ -605,7 +621,7 @@ void smbios_get_info(hd_data_t *hd_data, memory_range_t *mem, bios_info_t *bt)
       len = *(unsigned short *) (mem->data + u + 0x16);
       structs = *(unsigned short *) (mem->data + u + 0x1c);
       if(hlen < 0x1e) continue;
-      ok = crc(mem->data + u, hlen) == 0 && addr < (1 << 20) && len;
+      ok = crc(mem->data + u, hlen) == 0 && len;
       if(ok) break;
     }
   }
@@ -624,7 +640,7 @@ void smbios_get_info(hd_data_t *hd_data, memory_range_t *mem, bios_info_t *bt)
   memory.start = addr;
   memory.size = len;
   memory.data = NULL;
-  read_memory(&memory);
+  read_memory(hd_data, &memory);
   if(len >= 0x4000) {
     ADD2LOG(
       "  SMBIOS Structure Table at 0x%05x (size 0x%x)\n",
@@ -767,9 +783,10 @@ void get_fsc_info(hd_data_t *hd_data, memory_range_t *mem, bios_info_t *bt)
 }
 
 
+#ifndef LIBHD_TINY
 void add_panel_info(hd_data_t *hd_data, bios_info_t *bt)
 {
-  unsigned width, height;
+  unsigned width, height, xsize = 0, ysize = 0;
   char *vendor, *name, *version;
   hd_smbios_t *sm;
   unsigned u;
@@ -790,6 +807,20 @@ void add_panel_info(hd_data_t *hd_data, bios_info_t *bt)
 
   if(!vendor || !name) return;
 
+  if(
+    !strcmp(vendor, "IBM") &&
+    tp_lookup(name, &width, &height, &xsize, &ysize)
+  ) {
+    bt->lcd.vendor = new_str(vendor);
+    bt->lcd.name = new_str("Notebook LCD");
+    bt->lcd.width = width;
+    bt->lcd.height = height;
+    bt->lcd.xsize = xsize;
+    bt->lcd.ysize = ysize;
+
+    return;
+  }
+
   for(u = 0; u < sizeof panel_data / sizeof *panel_data; u++) {
     if(
       !strcmp(vendor, panel_data[u].vendor) &&
@@ -801,11 +832,13 @@ void add_panel_info(hd_data_t *hd_data, bios_info_t *bt)
       bt->lcd.name = new_str("Notebook LCD");
       bt->lcd.width = panel_data[u].width;
       bt->lcd.height = panel_data[u].height;
+      bt->lcd.xsize = panel_data[u].xsize;
+      bt->lcd.ysize = panel_data[u].ysize;
       break;
     }
   }
 }
-
+#endif
 
 void add_mouse_info(hd_data_t *hd_data, bios_info_t *bt)
 {
@@ -877,7 +910,7 @@ int get_smp_info(hd_data_t *hd_data, memory_range_t *mem, smp_info_t *smp)
     if(*(unsigned *) (mem->data + u) == 0x5f504d5f) {	/* "_MP_" */
       addr = *(unsigned *) (mem->data + u + 4);
       len = mem->data[u + 8];
-      ok = len == 1 && crc(mem->data + u, 0x10) == 0 && addr < (1 << 20) ? 1 : 0;
+      ok = len == 1 && crc(mem->data + u, 0x10) == 0 ? 1 : 0;
       ADD2LOG(
         "  smp: %svalid MP FP at 0x%05x (size 0x%x, rev %u), MP config at 0x%05x\n",
         ok ? "" : "in", u + mem->start, len << 4, mem->data[u + 9], addr
@@ -1002,8 +1035,56 @@ int get_bios32_info(hd_data_t *hd_data, memory_range_t *mem, bios32_info_t *bios
   return ok;
 }
 
+/*
+ *
+ * db format (32 bits):
+ * leaf: 1, last: 1, key: 6, range: 4, ofs: 20
+ *
+ */
+int tp_lookup(char *key_str, unsigned *width, unsigned *height, unsigned *xsize, unsigned *ysize)
+{
+  unsigned u;
+  unsigned key, range, ofs, last, leaf = 0;
 
-#endif		/* !defined(LIBHD_TINY) */
+  if(strlen(key_str) != 7) return 0;
 
+  for(u = 0; u < 7; u++) {
+    if(key_str[u] < '0' || key_str[u] >= '0' + 64) return 0;
+  }
+
+  for(u = 0; u < sizeof tp_db / sizeof *tp_db; u++) {
+    key = (tp_db[u] >> 24) & 0x3f;
+    range = (tp_db[u] >> 20) & 0xf;
+    ofs = tp_db[u] & ((1 << 20) - 1);
+    // printf("key = %d, range = %d, ofs = %d, str = %d\n", key, range, ofs, *key_str - '0');
+    if(*key_str - '0' >= key && *key_str - '0' <= key + range) {
+      // printf("match\n");
+      leaf = (tp_db[u] >> 31) & 1;
+      if(leaf) break;
+      key_str++;
+      // printf("%d\n", *key_str);
+      if(ofs >= sizeof tp_db / sizeof *tp_db || !*key_str) return 0;
+      u = ofs - 1;
+      // printf("next1 = %u\n", ofs);
+      continue;
+    }
+    else {
+      last = (tp_db[u] >> 30) & 1;
+      if(last) return 0;
+      // printf("next2 = %u\n", u + 1);
+    }
+  }
+
+  if(leaf) {
+    if(ofs >= sizeof tp_values / sizeof *tp_values) return 0;
+    *width = tp_values[ofs].width;
+    *height = tp_values[ofs].height;
+    *xsize = tp_values[ofs].xsize;
+    *ysize = tp_values[ofs].ysize;
+  }
+
+  return 1;
+}
 
 #endif /* defined(__i386__) || defined (__x86_64__) */
+
