@@ -9,16 +9,13 @@
 #include <sys/types.h>
 #include <sys/ioctl.h>
 #include <linux/iso_fs.h>
+#include <scsi/sg.h>
 
 #include "hd.h"
 #include "hd_int.h"
 #include "hddb.h"
 #include "block.h"
 
-#ifndef SCSI_IOCTL_SEND_COMMAND
-#define SCSI_IOCTL_SEND_COMMAND 1
-#endif
-
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  * block device stuff
  *
@@ -37,6 +34,7 @@ static cdrom_info_t *new_cdrom_entry(cdrom_info_t **ci);
 static cdrom_info_t *get_cdrom_entry(cdrom_info_t *ci, int n);
 static void get_scsi_tape(hd_data_t *hd_data);
 static void get_generic_scsi_devs(hd_data_t *hd_data);
+static void add_disk_size(hd_data_t *hd_data, hd_t *hd);
 
 
 void hd_scan_sysfs_block(hd_data_t *hd_data)
@@ -156,6 +154,7 @@ void get_block_devs(hd_data_t *hd_data)
 
     hd = NULL;
 
+#if defined(__s390x__) || defined(__s390__)
     /* check if disk is DASD and has already been found by s390.c */
     if(sf_dev && sf_dev->driver_name && strstr(sf_dev->driver_name,"dasd"))
     {
@@ -186,7 +185,9 @@ void get_block_devs(hd_data_t *hd_data)
       hd=NULL;
       out:;
     }
-    else if((sl = search_str_list(hd_data->disks, hd_sysfs_name2_dev(sf_cdev->name)))) {
+    else
+#endif
+    if((sl = search_str_list(hd_data->disks, hd_sysfs_name2_dev(sf_cdev->name)))) {
       hd = add_hd_entry(hd_data, __LINE__, 0);
       hd->sub_class.id = sc_sdev_disk;
     }
@@ -194,7 +195,11 @@ void get_block_devs(hd_data_t *hd_data)
       hd = add_hd_entry(hd_data, __LINE__, 0);
       hd->sub_class.id = sc_sdev_cdrom;
     }
-    else if(sf_dev && sf_dev->bus && !strcmp(sf_dev->bus, "scsi")) {
+    else if(
+      sf_dev &&
+      sf_dev->bus &&
+      (!strcmp(sf_dev->bus, "scsi") || !strcmp(sf_dev->bus, "ide"))
+    ) {
       hd = add_hd_entry(hd_data, __LINE__, 0);
       hd->sub_class.id = sc_sdev_other;
     }
@@ -258,6 +263,18 @@ void get_block_devs(hd_data_t *hd_data)
         }
       }
 
+      /*
+       * set hd->drivers before calling any of add_xxx_sysfs_info()
+       */
+      if(
+        sf_dev &&
+        sf_dev->driver_name &&
+        *sf_dev->driver_name &&
+        strcmp(sf_dev->driver_name, "unknown")
+      ) {
+        add_str_list(&hd->drivers, sf_dev->driver_name);
+      }
+
       if(hd->bus.id == bus_ide) {
         add_ide_sysfs_info(hd_data, hd, sf_dev);
       }
@@ -269,15 +286,6 @@ void get_block_devs(hd_data_t *hd_data)
       }
 
       
-      if(
-        sf_dev &&
-        sf_dev->driver_name &&
-        *sf_dev->driver_name &&
-        strcmp(sf_dev->driver_name, "unknown")
-      ) {
-        add_str_list(&hd->drivers, sf_dev->driver_name);
-      }
-
       if(hd->sub_class.id == sc_sdev_cdrom) {
         add_cdrom_info(hd_data, hd);
       }
@@ -377,43 +385,43 @@ void add_cdrom_info(hd_data_t *hd_data, hd_t *hd)
     }
   }
 
-  ci = hd->detail->cdrom.data;
-
-  /* update prog_if: cdr, cdrw, ... */
-  if(
-    /* ###### FIXME: dosn't work anyway: ide-scsi doesn't support sysfs */
-    hd->bus.id == bus_scsi &&
-    !search_str_list(hd->drivers, "ide-scsi")		/* could be ide, though */
-  ) {
-    /* scsi devs lie */
-    if(ci->dvd && (ci->cdrw || ci->dvdr || ci->dvdram)) {
-      ci->dvd = ci->dvdr = ci->dvdram = 0;
+  if((ci = hd->detail->cdrom.data)) {
+    /* update prog_if: cdr, cdrw, ... */
+    if(
+      /* ###### FIXME: dosn't work anyway: ide-scsi doesn't support sysfs */
+      hd->bus.id == bus_scsi &&
+      !search_str_list(hd->drivers, "ide-scsi")		/* could be ide, though */
+    ) {
+      /* scsi devs lie */
+      if(ci->dvd && (ci->cdrw || ci->dvdr || ci->dvdram)) {
+        ci->dvd = ci->dvdr = ci->dvdram = 0;
+      }
+      ci->dvdr = ci->dvdram = 0;
+      ci->cdr = ci->cdrw = 0;
+      if(hd->prog_if.id == pif_cdr) ci->cdr = 1;
     }
-    ci->dvdr = ci->dvdram = 0;
-    ci->cdr = ci->cdrw = 0;
-    if(hd->prog_if.id == pif_cdr) ci->cdr = 1;
-  }
 
-  /* trust ide info */
-  if(ci->dvd) {
-    hd->is.dvd = 1;
-    hd->prog_if.id = pif_dvd;
-  }
-  if(ci->cdr) {
-    hd->is.cdr = 1;
-    hd->prog_if.id = pif_cdr;
-  }
-  if(ci->cdrw) {
-    hd->is.cdrw = 1;
-    hd->prog_if.id = pif_cdrw;
-  }
-  if(ci->dvdr) {
-    hd->is.dvdr = 1;
-    hd->prog_if.id = pif_dvdr;
-  }
-  if(ci->dvdram) {
-    hd->is.dvdram = 1;
-    hd->prog_if.id = pif_dvdram;
+    /* trust ide info */
+    if(ci->dvd) {
+      hd->is.dvd = 1;
+      hd->prog_if.id = pif_dvd;
+    }
+    if(ci->cdr) {
+      hd->is.cdr = 1;
+      hd->prog_if.id = pif_cdr;
+    }
+    if(ci->cdrw) {
+      hd->is.cdrw = 1;
+      hd->prog_if.id = pif_cdrw;
+    }
+    if(ci->dvdr) {
+      hd->is.dvdr = 1;
+      hd->prog_if.id = pif_dvdr;
+    }
+    if(ci->dvdram) {
+      hd->is.dvdram = 1;
+      hd->prog_if.id = pif_dvdram;
+    }
   }
 
   if(
@@ -427,10 +435,8 @@ void add_cdrom_info(hd_data_t *hd_data, hd_t *hd)
 
 void add_other_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_dev)
 {
-  hd_res_t *geo, *size;
-  int fd;
   unsigned u0, u1;
-  char c, *pr_str;
+  char c;
 
   if(hd->sysfs_id) {
     if(
@@ -467,41 +473,20 @@ void add_other_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_
     }
   }
 
-
-  pr_str = NULL;
-
-  if(
-    hd->unix_dev_name &&
-    hd->sub_class.id == sc_sdev_disk
-  ) {
-    PROGRESS(5, 0, hd->unix_dev_name);
-    fd = open(hd->unix_dev_name, O_RDONLY | O_NONBLOCK);
-    if(fd >= 0) {
-
-      str_printf(&pr_str, 0, "%s geo", hd->unix_dev_name);
-      PROGRESS(5, 1, pr_str);
-
-      hd_getdisksize(hd_data, hd->unix_dev_name, fd, &geo, &size);
-      if(geo) add_res_entry(&hd->res, geo);
-      if(size) add_res_entry(&hd->res, size);
-
-      close(fd);
-    }
-  }
-
-  pr_str = free_mem(pr_str);
-
+  add_disk_size(hd_data, hd);
 }
 
 
 void add_ide_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_dev)
 {
   char *fname = NULL, buf[256], *dev_name, *s;
-  unsigned u0, u1, u2;
+  unsigned u0, u1, u2, size = 0;
   str_list_t *sl, *sl0;
   hd_res_t *res;
   FILE *f;
 
+  if(!hd_report_this(hd_data, hd)) return;
+
   if(hd->sysfs_bus_id && sscanf(hd->sysfs_bus_id, "%u.%u", &u0, &u1) == 2) {
     /* host.master/slave */
     hd->slot = (u0 << 1) + u1;
@@ -555,45 +540,51 @@ void add_ide_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_de
       free_str_list(sl);
     }
 
-
-    str_printf(&fname, 0, PROC_IDE "/%s/geometry", dev_name);
-    if((sl0 = read_file(fname, 0, 2))) {
-      for(sl = sl0; sl; sl = sl->next) {
-        if(sscanf(sl->str, " physical %u / %u / %u", &u0, &u1, &u2) == 3) {
-          if(u0 || u1 || u2) {
-            res = add_res_entry(&hd->res, new_mem(sizeof *res));
-            res->disk_geo.type = res_disk_geo;
-            res->disk_geo.cyls = u0;
-            res->disk_geo.heads = u1;
-            res->disk_geo.sectors = u2;
-          }
-          continue;
-        }
-
-        if(sscanf(sl->str, " logical %u / %u / %u", &u0, &u1, &u2) == 3) {
-          res = add_res_entry(&hd->res, new_mem(sizeof *res));
-          res->disk_geo.type = res_disk_geo;
-          res->disk_geo.cyls = u0;
-          res->disk_geo.heads = u1;
-          res->disk_geo.sectors = u2;
-          res->disk_geo.logical = 1;
-        }
-      }
-      free_str_list(sl0);
-    }
-
     str_printf(&fname, 0, PROC_IDE "/%s/capacity", dev_name);
     if((sl = read_file(fname, 0, 1))) {
       if(sscanf(sl->str, "%u", &u0) == 1 && u0 != 0x7fffffff) {
         res = add_res_entry(&hd->res, new_mem(sizeof *res));
         res->size.type = res_size;
         res->size.unit = size_unit_sectors;
-        res->size.val1 = u0;
+        res->size.val1 = size = u0;
         res->size.val2 = 512;		// ####### FIXME: sector size?
       }
       free_str_list(sl);
     }
 
+    str_printf(&fname, 0, PROC_IDE "/%s/geometry", dev_name);
+    if((sl0 = read_file(fname, 0, 2))) {
+      for(sl = sl0; sl; sl = sl->next) {
+        if(sscanf(sl->str, " physical %u / %u / %u", &u0, &u1, &u2) == 3) {
+          if(u0 || u1 || u2) {
+            if(size && u1 && u2) {
+              u0 = size / (u1 * u2);
+            }
+            res = add_res_entry(&hd->res, new_mem(sizeof *res));
+            res->disk_geo.type = res_disk_geo;
+            res->disk_geo.cyls = u0;
+            res->disk_geo.heads = u1;
+            res->disk_geo.sectors = u2;
+            res->disk_geo.geotype = geo_physical;
+          }
+          continue;
+        }
+
+        if(sscanf(sl->str, " logical %u / %u / %u", &u0, &u1, &u2) == 3) {
+          if(size && u1 && u2) {
+            u0 = size / (u1 * u2);
+          }
+          res = add_res_entry(&hd->res, new_mem(sizeof *res));
+          res->disk_geo.type = res_disk_geo;
+          res->disk_geo.cyls = u0;
+          res->disk_geo.heads = u1;
+          res->disk_geo.sectors = u2;
+          res->disk_geo.geotype = geo_logical;
+        }
+      }
+      free_str_list(sl0);
+    }
+
     str_printf(&fname, 0, PROC_IDE "/%s/cache", dev_name);
     if((sl = read_file(fname, 0, 1))) {
       if(sscanf(sl->str, "%u", &u0) == 1 && u0) {
@@ -625,9 +616,14 @@ void add_ide_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_de
 
     free_mem(fname);
   }
+
+  if(!size) add_disk_size(hd_data, hd);
 }
 
 
+/*
+ * assumes hd->drivers aleady includes scsi device drivers (like 'sd')
+ */
 void add_scsi_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_dev)
 {
   hd_t *hd1;
@@ -635,11 +631,14 @@ void add_scsi_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_d
   unsigned u0, u1, u2, u3;
   int fd, k;
   unsigned char scsi_cmd_buf[0x300];
+  struct sg_io_hdr hdr;
   unsigned char *uc;
   scsi_t *scsi;
   hd_res_t *geo, *size;
   uint64_t ul0;
 
+  if(!hd_report_this(hd_data, hd)) return;
+
   hd->detail = new_mem(sizeof *hd->detail);
   hd->detail->type = hd_detail_scsi;
   hd->detail->scsi.data = scsi = new_mem(sizeof *scsi);
@@ -670,6 +669,32 @@ void add_scsi_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_d
     else {
       free_mem(cs);
     }
+
+    /* sata entries are somewhat strange... */
+    if(
+      hd->vendor.name &&
+      !strcmp(hd->vendor.name, "ATA") &&
+      hd->device.name
+    ) {
+      hd->bus.id = bus_ide;
+
+      if((cs = strchr(hd->device.name, ' '))) {
+        t = canon_str(cs, strlen(cs));
+        if(*t) {
+          *cs = 0;
+          free_mem(hd->vendor.name);
+          hd->vendor.name = hd->device.name;
+          hd->device.name = t;
+        }
+        else {
+          t = free_mem(t);
+        }
+      }
+
+      if(!strcmp(hd->vendor.name, "ATA")) {
+        hd->vendor.name = free_mem(hd->vendor.name);
+      }
+    }
   }
 
   if((s = hd_attr_str(sysfs_get_device_attr(sf_dev, "rev")))) {
@@ -697,17 +722,24 @@ void add_scsi_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_d
       }
     }
 
+    /*
+     * typically needed for usb card readers (unused slots)
+     */
     if(
       hd->base_class.id == bc_storage_device &&
       hd->sub_class.id == sc_sdev_other
     ) {
       switch(ul0) {
         case 0:
-          hd->sub_class.id = sc_sdev_disk;
+          if(search_str_list(hd->drivers, "sd")) {
+            hd->sub_class.id = sc_sdev_disk;
+          }
           break;
 
         case 5:
-          hd->sub_class.id = sc_sdev_cdrom;
+          if(search_str_list(hd->drivers, "sr")) {
+            hd->sub_class.id = sc_sdev_cdrom;
+          }
           break;
       }
     }
@@ -757,19 +789,27 @@ void add_scsi_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_d
       PROGRESS(5, 1, pr_str);
 
       memset(scsi_cmd_buf, 0, sizeof scsi_cmd_buf);
-      // ###### FIXME: smaller!
-      *((unsigned *) (scsi_cmd_buf + 4)) = sizeof scsi_cmd_buf - 0x100;
-      scsi_cmd_buf[8 + 0] = 0x1a;
-      scsi_cmd_buf[8 + 2] = 0x08;
-      scsi_cmd_buf[8 + 4] = 0xff;
+      memset(&hdr, 0, sizeof(hdr));
 
-      k = ioctl(fd, SCSI_IOCTL_SEND_COMMAND, scsi_cmd_buf);
+      hdr.interface_id = 'S';
+      hdr.cmd_len = 6;
+      hdr.dxfer_direction = SG_DXFER_FROM_DEV;
+      hdr.dxferp = scsi_cmd_buf + 8 + 6;
+      hdr.dxfer_len = 0xff;
+      hdr.cmdp = scsi_cmd_buf + 8;
+      hdr.cmdp[0] = 0x1a;
+      hdr.cmdp[2] = 0x08;
+      hdr.cmdp[4] = 0xff;
+
+      k = ioctl(fd, SG_IO, &hdr);
 
       if(k) {
         ADD2LOG("%s status(0x1a:8) 0x%x\n", hd->unix_dev_name, k);
       }
       else {
-        uc = scsi_cmd_buf + 8 + 4 + scsi_cmd_buf[8 + 3] + 2;
+        unsigned char *ptr = hdr.dxferp;
+
+        uc = ptr + 4 + ptr[3] + 2;
         scsi->cache = uc[0];
         ADD2LOG("  scsi cache: 0x%02x\n", uc[0]);
     
@@ -786,7 +826,8 @@ void add_scsi_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_d
   if(
     hd_report_this(hd_data, hd) &&
     hd->unix_dev_name &&
-    hd->sub_class.id == sc_sdev_disk
+    hd->sub_class.id == sc_sdev_disk &&
+    !hd_probe_feature(hd_data, pr_scsi_noserial)
   ) {
     PROGRESS(5, 0, hd->unix_dev_name);
     fd = open(hd->unix_dev_name, O_RDONLY | O_NONBLOCK);
@@ -795,7 +836,11 @@ void add_scsi_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_d
       str_printf(&pr_str, 0, "%s geo", hd->unix_dev_name);
       PROGRESS(5, 1, pr_str);
 
-      hd_getdisksize(hd_data, hd->unix_dev_name, fd, &geo, &size);
+      if(hd_getdisksize(hd_data, hd->unix_dev_name, fd, &geo, &size) == 1) {
+        /* (low-level) unformatted disk */
+        hd->is.notready = 1;
+      }
+
       if(geo) add_res_entry(&hd->res, geo);
       if(size) add_res_entry(&hd->res, size);
 
@@ -803,20 +848,30 @@ void add_scsi_sysfs_info(hd_data_t *hd_data, hd_t *hd, struct sysfs_device *sf_d
       PROGRESS(5, 2, pr_str);
 
       memset(scsi_cmd_buf, 0, sizeof scsi_cmd_buf);
-      // ###### FIXME: smaller!
-      *((unsigned *) (scsi_cmd_buf + 4)) = sizeof scsi_cmd_buf - 0x100;
-      scsi_cmd_buf[8 + 0] = 0x12;
-      scsi_cmd_buf[8 + 1] = 0x01;
-      scsi_cmd_buf[8 + 2] = 0x80;
-      scsi_cmd_buf[8 + 4] = 0xff;
+      memset(&hdr, 0, sizeof(hdr));
 
-      k = ioctl(fd, SCSI_IOCTL_SEND_COMMAND, scsi_cmd_buf);
+      hdr.interface_id = 'S';
+      hdr.cmd_len = 6;
+      hdr.dxfer_direction = SG_DXFER_FROM_DEV;
+      hdr.dxferp = scsi_cmd_buf + 8 + 6;
+      hdr.dxfer_len = 0x24;
+      hdr.cmdp = scsi_cmd_buf + 8;
+      hdr.cmdp[0] = 0x12;
+      hdr.cmdp[1] = 0x01;
+      hdr.cmdp[2] = 0x80;
+      hdr.cmdp[4] = 0x24;
+ 
+      k = ioctl(fd, SG_IO, &hdr);
 
       if(k) {
         ADD2LOG("%s status(0x12) 0x%x\n", scsi->dev_name, k);
       }
       else {
-        if((hd->serial = canon_str(scsi_cmd_buf + 8 + 4, scsi_cmd_buf[8 + 3]))) {
+        unsigned char *ptr = hdr.dxferp;
+
+        ADD2LOG("  serial id len: %u\n", ptr[3]);
+
+        if((hd->serial = canon_str(ptr + 4, ptr[3]))) {
           if(!*hd->serial) hd->serial = free_mem(hd->serial);
         }
       }
@@ -1218,11 +1273,15 @@ void hd_scan_sysfs_scsi(hd_data_t *hd_data)
   /* some clean-up */
   remove_hd_entries(hd_data);
 
-  PROGRESS(1, 0, "scsi tape");
+  PROGRESS(1, 0, "scsi modules");
+
+  load_module(hd_data, "sg");
+
+  PROGRESS(2, 0, "scsi tape");
 
   get_scsi_tape(hd_data);
 
-  PROGRESS(2, 0, "scsi generic");
+  PROGRESS(3, 0, "scsi generic");
 
   get_generic_scsi_devs(hd_data);
 }
@@ -1485,3 +1544,38 @@ void get_generic_scsi_devs(hd_data_t *hd_data)
 }
 
 
+void add_disk_size(hd_data_t *hd_data, hd_t *hd)
+{
+  hd_res_t *geo, *size;
+  int fd;
+  char *pr_str;
+
+  pr_str = NULL;
+
+  if(
+    hd->unix_dev_name &&
+    hd->sub_class.id == sc_sdev_disk
+  ) {
+    PROGRESS(5, 0, hd->unix_dev_name);
+    fd = open(hd->unix_dev_name, O_RDONLY | O_NONBLOCK);
+    if(fd >= 0) {
+
+      str_printf(&pr_str, 0, "%s geo", hd->unix_dev_name);
+      PROGRESS(5, 1, pr_str);
+
+      if(hd_getdisksize(hd_data, hd->unix_dev_name, fd, &geo, &size) == 1) {
+        /* (low-level) unformatted disk */
+        hd->is.notready = 1;
+      }
+
+      if(geo) add_res_entry(&hd->res, geo);
+      if(size) add_res_entry(&hd->res, size);
+
+      close(fd);
+    }
+  }
+
+  pr_str = free_mem(pr_str);
+}
+
+
