@@ -13,7 +13,7 @@ extern "C" {
  * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  */
 
-#define HD_VERSION	8
+#define HD_VERSION	12
 
 /*
  * debug flags
@@ -46,12 +46,12 @@ extern "C" {
 #include <stdio.h>
 #include <inttypes.h>
 #include <termios.h>
+#include <sys/types.h>
 
 /*
  * libhd's directory
  */
 #define HARDWARE_DIR		"/var/lib/hardware"
-#define HARDWARE_UNIQUE_KEYS	HARDWARE_DIR "/unique-keys"
 
 /**
  * \defgroup idmacros Id macros
@@ -105,7 +105,8 @@ typedef enum probe_feature {
   pr_parallel_imm, pr_s390, pr_cpuemu, pr_sysfs, pr_s390disks, pr_udev,
   pr_block, pr_block_cdrom, pr_block_part, pr_edd, pr_edd_mod, pr_bios_ddc,
   pr_bios_fb, pr_bios_mode, pr_input, pr_block_mods, pr_bios_vesa,
-  pr_max, pr_lxrc, pr_dsl, pr_default, pr_all		/* pr_all must be last */
+  pr_cpuemu_debug, pr_scsi_noserial, pr_wlan, pr_bios_crc, pr_hal,
+  pr_max, pr_lxrc, pr_default, pr_all		/* pr_all must be last */
 } hd_probe_feature_t;
 
 /*
@@ -114,16 +115,19 @@ typedef enum probe_feature {
  * if you want to modify this: cf. manual.c::hw_items[]
  *
  * Note: hw_tv _must_ be < hw_display!
+ * Sync with check_hd and convert_hd!
  */
 typedef enum hw_item {
-  hw_sys = 1, hw_cpu, hw_keyboard, hw_braille, hw_mouse, hw_joystick,
-  hw_printer, hw_scanner, hw_chipcard, hw_monitor, hw_tv, hw_display,
-  hw_framebuffer, hw_camera, hw_sound, hw_storage_ctrl, hw_network_ctrl,
-  hw_isdn, hw_modem, hw_network, hw_disk, hw_partition, hw_cdrom, hw_floppy,
-  hw_manual, hw_usb_ctrl, hw_usb, hw_bios, hw_pci, hw_isapnp, hw_bridge,
-  hw_hub, hw_scsi, hw_ide, hw_memory, hw_dvb, hw_pcmcia, hw_pcmcia_ctrl,
-  hw_ieee1394, hw_ieee1394_ctrl, hw_hotplug, hw_hotplug_ctrl, hw_zip, hw_pppoe,
-  hw_wlan, hw_redasd, hw_dsl, hw_block, hw_tape, hw_vbe, /* append new entries here */
+  hw_none = 0, hw_sys, hw_cpu, hw_keyboard, hw_braille, hw_mouse,
+  hw_joystick, hw_printer, hw_scanner, hw_chipcard, hw_monitor, hw_tv,
+  hw_display, hw_framebuffer, hw_camera, hw_sound, hw_storage_ctrl,
+  hw_network_ctrl, hw_isdn, hw_modem, hw_network, hw_disk, hw_partition,
+  hw_cdrom, hw_floppy, hw_manual, hw_usb_ctrl, hw_usb, hw_bios, hw_pci,
+  hw_isapnp, hw_bridge, hw_hub, hw_scsi, hw_ide, hw_memory, hw_dvb,
+  hw_pcmcia, hw_pcmcia_ctrl, hw_ieee1394, hw_ieee1394_ctrl, hw_hotplug,
+  hw_hotplug_ctrl, hw_zip, hw_pppoe, hw_wlan, hw_redasd, hw_dsl, hw_block,
+  hw_tape, hw_vbe, hw_bluetooth,
+  /* append new entries here */
   hw_unknown, hw_all					/* hw_all must be last */
 } hd_hw_item_t;
 
@@ -143,7 +147,7 @@ typedef enum base_classes {
   bc_monitor = 0x100, bc_internal, bc_modem, bc_isdn, bc_ps2, bc_mouse,
   bc_storage_device, bc_network_interface, bc_keyboard, bc_printer,
   bc_hub, bc_braille, bc_scanner, bc_joystick, bc_chipcard, bc_camera,
-  bc_framebuffer, bc_dvb, bc_tv, bc_partition, bc_dsl
+  bc_framebuffer, bc_dvb, bc_tv, bc_partition, bc_dsl, bc_bluetooth
 } hd_base_classes_t;
 
 /* subclass values of bc_monitor */
@@ -217,7 +221,8 @@ typedef enum sc_std {
 typedef enum sc_net_if {
   sc_nif_loopback, sc_nif_ethernet, sc_nif_tokenring, sc_nif_fddi,
   sc_nif_ctc, sc_nif_iucv, sc_nif_hsi, sc_nif_qeth,
-  sc_nif_escon, sc_nif_myrinet, sc_nif_wlan, sc_nif_other = 0x80, sc_nif_sit
+  sc_nif_escon, sc_nif_myrinet, sc_nif_wlan, sc_nif_xp,
+  sc_nif_usb, sc_nif_other = 0x80, sc_nif_sit
 } hd_sc_net_if_t;
 
 /* subclass values of bc_multimedia */
@@ -263,7 +268,7 @@ typedef enum pif_cdrom {
 
 /* S/390 disk prog_if values */
 typedef enum pif_s390disk {
-  pif_scsi, pif_dasd
+  pif_scsi, pif_dasd, pif_dasd_fba
 } hd_pif_s390disk_t;
 
 /* bus type values similar to PCI bridge subclasses */
@@ -273,7 +278,7 @@ typedef enum bus_types {
 
   /* outside the range of the PCI values */
   bus_ps2 = 0x80, bus_serial, bus_parallel, bus_floppy, bus_scsi, bus_ide, bus_usb,
-  bus_adb, bus_raid, bus_sbus, bus_i2o, bus_vio
+  bus_adb, bus_raid, bus_sbus, bus_i2o, bus_vio, bus_ccw, bus_iucv
 } hd_bus_types_t;
 
 /**
@@ -444,7 +449,7 @@ typedef struct {
   unsigned modes;		/* number of supported video modes */
   vbe_mode_info_t *mode;	/* video mode list */
   unsigned current_mode;	/* current video mode */
-  unsigned char ddc[0x80];	/* ddc monitor info */
+  unsigned char ddc_port[4][0x80];	/* ddc monitor info per port */
 } vbe_info_t;
 
 
@@ -973,6 +978,7 @@ typedef struct s_pci_t {
   uint64_t rom_base_len;			/* memory range for card ROM */
   char *sysfs_id;				/* sysfs path */
   char *sysfs_bus_id;				/* sysfs bus id */
+  char *modalias;				/* module alias */
 } pci_t;
 
 /*
@@ -1171,6 +1177,8 @@ typedef struct {
   struct {
     unsigned width;
     unsigned height;
+    unsigned xsize;
+    unsigned ysize;
     char *vendor;
     char *name;
   } lcd;
@@ -1214,6 +1222,7 @@ typedef struct {
   char *model;
   char *serial;
   char *lang;
+  char *formfactor;
 } sys_info_t;
 
 
@@ -1224,6 +1233,9 @@ typedef struct {
   unsigned manu_year;
   unsigned min_vsync, max_vsync;	/* vsync range */
   unsigned min_hsync, max_hsync;	/* hsync range */
+  unsigned clock;			/* pixel clock in kHz */
+  unsigned width, height;		/* display size */
+  unsigned width_mm, height_mm;		/* dto, in mm */
   char *vendor;
   char *name;
   char *serial;
@@ -1267,6 +1279,27 @@ typedef struct {
 } cpu_info_t;
 
 
+/*
+ * enhanced disk data (cf. edd.c)
+ */
+typedef struct {
+  uint64_t sectors;
+  struct {
+    unsigned cyls, heads, sectors;
+  } edd;
+  struct {
+    unsigned cyls, heads, sectors;
+  } legacy;
+  unsigned ext_fixed_disk:1;
+  unsigned ext_lock_eject:1;
+  unsigned ext_edd:1;
+  unsigned ext_64bit:1;
+  unsigned assigned:1;
+  char *sysfs_id;
+  unsigned hd_idx;
+} edd_info_t;
+
+
 /*
  * database info
  */
@@ -1296,17 +1329,85 @@ typedef struct {
 } hddb2_data_t;
 
 /*
- * pci module info
+ * module.alias info
  */
+typedef enum modinfo_type_e { mi_none = 0, mi_pci, mi_usb, mi_pcmcia } modinfo_type_t;
+
 typedef struct {
   char *module;
-  unsigned vendor;
-  unsigned device;
-  unsigned subvendor;
-  unsigned subdevice;
-  unsigned pciclass;
-  unsigned classmask;
-} hddb_pci_t;
+  char *alias;
+  modinfo_type_t type;
+  union {
+    struct {
+      struct {
+        unsigned vendor:1;
+        unsigned device:1;
+        unsigned sub_vendor:1;
+        unsigned sub_device:1;
+        unsigned base_class:1;
+        unsigned sub_class:1;
+        unsigned prog_if:1;
+      } has;
+      unsigned vendor;
+      unsigned device;
+      unsigned sub_vendor;
+      unsigned sub_device;
+      unsigned base_class;
+      unsigned sub_class;
+      unsigned prog_if;
+    } pci;
+
+    struct {
+      struct {
+        unsigned vendor:1;
+        unsigned product:1;
+        unsigned device_class:1;
+        unsigned device_subclass:1;
+      } has;
+      unsigned vendor;
+      unsigned product;
+      unsigned device_class;
+      unsigned device_subclass;
+    } usb;
+  };
+} modinfo_t;
+
+
+/*
+ * HAL device property types
+ */
+typedef enum {
+  p_invalid, p_string, p_int32, p_uint64, p_double, p_bool, p_list
+} hal_prop_type_t;
+
+
+/*
+ * HAL device properties
+ */
+typedef struct hal_prop_s {
+  struct hal_prop_s *next;
+  hal_prop_type_t type;
+  char *key;
+  union {
+    char *str;
+    int32_t int32;
+    uint64_t uint64;
+    double d;     
+    int b;
+    str_list_t *list;
+  } val;  
+} hal_prop_t;
+
+
+/*
+ * HAL device
+ */
+typedef struct hal_device_s {
+  struct hal_device_s *next, *parent;
+  char *udi;
+  unsigned used:1;
+  hal_prop_t *prop;
+} hal_device_t;
 
 
 /*
@@ -1315,7 +1416,7 @@ typedef struct {
 typedef enum resource_types {
   res_any, res_phys_mem, res_mem, res_io, res_irq, res_dma, res_monitor,
   res_size, res_disk_geo, res_cache, res_baud, res_init_strings, res_pppd_option,
-  res_framebuffer, res_hwaddr
+  res_framebuffer, res_hwaddr, res_link, res_wlan
 } hd_resource_types_t;
 
 
@@ -1324,7 +1425,7 @@ typedef enum resource_types {
  */
 typedef enum size_units {
   size_unit_cm, size_unit_cinch, size_unit_byte, size_unit_sectors,
-  size_unit_kbyte, size_unit_mbyte, size_unit_gbyte
+  size_unit_kbyte, size_unit_mbyte, size_unit_gbyte, size_unit_mm
 } hd_size_units_t;
 
 /*
@@ -1334,11 +1435,17 @@ typedef enum access_flags {
   acc_unknown, acc_ro, acc_wo, acc_rw		/* unknown, read only, write only, read/write */
 } hd_access_flags_t;
 
+
 typedef enum yes_no_flag {
   flag_unknown, flag_no, flag_yes		/* unknown, no, yes */
 } hd_yes_no_flag_t;
 
 
+typedef enum geo_types {
+  geo_physical = 0, geo_logical, geo_bios_edd, geo_bios_legacy
+} hd_geo_types_t;
+
+
 /*
  * definitions for the various resource types
  */
@@ -1413,7 +1520,8 @@ typedef struct {
   union u_hd_res_t *next;
   enum resource_types type;
   unsigned cyls, heads, sectors;
-  unsigned logical:1;			/* logical/physical geometry */
+  uint64_t size;
+  enum geo_types geotype;		/* 0-3: physical/logical/bios edd/bios legacy */
 } res_disk_geo_t;
 
 typedef struct {
@@ -1452,6 +1560,23 @@ typedef struct {
   char *addr;
 } res_hwaddr_t;
 
+typedef struct {
+  union u_hd_res_t *next;
+  enum resource_types type;
+  unsigned state:1;			/* network link state: 0 - not connected, 1 - connected */
+} res_link_t;
+
+/* wlan capabilities */
+typedef struct {
+  union u_hd_res_t *next;
+  enum resource_types type;
+  str_list_t *channels;
+  str_list_t *frequencies; /* in GHz units */
+  str_list_t *bitrates;    /* in Mbps units */
+  str_list_t *auth_modes;  /* open, sharedkey, wpa-psk, wpa-eap, wpa-leap */
+  str_list_t *enc_modes;   /* WEP40, WEP104, WEP128, WEP232, TKIP, CCMP */
+} res_wlan_t;
+
 typedef union u_hd_res_t {
   union u_hd_res_t *next;  
   res_any_t any;
@@ -1469,6 +1594,8 @@ typedef union u_hd_res_t {
   res_pppd_option_t pppd_option;
   res_framebuffer_t framebuffer;
   res_hwaddr_t hwaddr;
+  res_link_t link;
+  res_wlan_t wlan;
 } hd_res_t;
 
 
@@ -1968,6 +2095,16 @@ typedef struct s_hd_t {
    */
   char *rom_id;
 
+  /**
+   * HAL udi.
+   */
+  char *udi;
+
+  /**
+   * \ref udi of parent (\ref attached_to).
+   */
+  char *parent_udi;
+
   /**
    * Unique id for this hardware.
    * A unique string identifying this hardware. The string consists
@@ -2057,6 +2194,7 @@ typedef struct s_hd_t {
     unsigned dvdram:1;		/* DVD-RAM */
     unsigned pppoe:1;		/* PPPOE modem connected */
     unsigned wlan:1;		/* WLAN card */
+    unsigned with_acpi:1;	/* acpi works fine */
   } is;
 
   struct tag_s {		/* this struct is for internal purposes only */
@@ -2125,6 +2263,12 @@ typedef struct s_hd_t {
 
   str_list_t *requires;		/* packages/programs required for this hardware */
 
+  hal_prop_t *hal_prop;		/* hal property list */
+
+  hal_prop_t *persistent_prop;	/* persistent property list */
+
+  char *modalias;		/* module alias */
+
   /*
    * These are used internally for memory management.
    * Do not even _think_ of modifying these!
@@ -2186,10 +2330,12 @@ typedef struct {
     unsigned cpuemu:1;		/**< use CPU emulation to run BIOS code (i386 only) */
     unsigned udev:1;		/**< return first udev symlink as device name */
     unsigned edd_used:1;	/**< internal: edd info has been used  */
+    unsigned keep_kmods:2;	/**< internal: don't reread kmods */
+    unsigned nobioscrc:1;	/**< internal: don't check VBIOS crc */
   } flags;
 
 
-  /** Concentrate on tses devices.
+  /** Concentrate on these devices.
    * List of sysfs ids for devices to look for.
    */
   str_list_t *only;
@@ -2218,8 +2364,8 @@ typedef struct {
   str_list_t *klog;		/**< (Internal) kernel log */
   str_list_t *proc_usb;		/**< (Internal) /proc/bus/usb info */
   usb_t *usb;			/**< (Internal) usb info */
-  hddb_pci_t *hddb_pci_hm;	/**< (Internal) pci module info */
-  hddb_pci_t *hddb_pci;		/**< (Internal) pci module info */
+  modinfo_t *modinfo_ext;	/**< (Internal) external module info */
+  modinfo_t *modinfo;		/**< (Internal) module info */
   hddb2_data_t *hddb2[2];	/**< (Internal) hardware database */
   str_list_t *kmods;		/**< (Internal) list of active kernel modules */
   uint64_t used_irqs;		/**< (Internal) irq usage */
@@ -2234,7 +2380,7 @@ typedef struct {
   devtree_t *devtree;		/**< (Internal) prom device tree on ppc */
   unsigned kernel_version;	/**< (Internal) kernel version */
   int in_vmware;		/**< (Internal) running in vmware */
-  hd_manual_t *manual;		/**< (Internal) hardware config info */
+  hd_t *manual;			/**< (Internal) hardware config info */
   str_list_t *disks;		/**< (Internal) disks according to /proc/partitions */
   str_list_t *partitions;	/**< (Internal) dto, partitions */
   str_list_t *cdroms;		/**< (Internal) cdroms according to PROC_CDROM_INFO */
@@ -2252,6 +2398,8 @@ typedef struct {
   hd_sysfsdrv_t *sysfsdrv;	/**< (Internal) sysfs driver info */
   uint64_t sysfsdrv_id;		/**< (Internal) sysfs driver info id */
   str_list_t *scanner_db;	/**< (Internal) list of scanner modules */
+  edd_info_t edd[0x80];		/**< (Internal) enhanced disk drive data */
+  hal_device_t *hal;		/**< (Internal) HAL data (if any) */
 } hd_data_t;
 
 
@@ -2286,14 +2434,15 @@ int hd_module_is_active(hd_data_t *hd_data, char *mod);
 hd_t *hd_base_class_list(hd_data_t *hd_data, unsigned base_class);
 hd_t *hd_sub_class_list(hd_data_t *hd_data, unsigned base_class, unsigned sub_class);
 hd_t *hd_bus_list(hd_data_t *hd_data, unsigned bus);
+const char* hd_busid_to_hwcfg(int busid);
 hd_t *hd_list(hd_data_t *hd_data, hd_hw_item_t item, int rescan, hd_t *hd_old);
 hd_t *hd_list_with_status(hd_data_t *hd_data, hd_hw_item_t item, hd_status_t status);
 hd_t *hd_list2(hd_data_t *hd_data, hd_hw_item_t *items, int rescan);
 hd_t *hd_list_with_status2(hd_data_t *hd_data, hd_hw_item_t *items, hd_status_t status);
 
-int hd_has_special_eide(hd_data_t *hd_data);
 int hd_has_pcmcia(hd_data_t *hd_data);
-int hd_apm_enabled(hd_data_t *hd_data);
+// will be gone soon
+// int hd_apm_enabled(hd_data_t *hd_data);
 int hd_usb_support(hd_data_t *hd_data);
 int hd_smp_support(hd_data_t *hd_data);
 int hd_mac_color(hd_data_t *hd_data);
@@ -2309,8 +2458,14 @@ hd_t *hd_get_device_by_idx(hd_data_t *hd_data, unsigned idx);
 void hd_set_hw_class(hd_t *hd, hd_hw_item_t hw_class);
 int hd_is_hw_class(hd_t *hd, hd_hw_item_t hw_class);
 
+int hd_is_sgi_altix(hd_data_t *hd_data);
+
 char *hd_version(void);
 
+hal_prop_t *hd_free_hal_properties(hal_prop_t *prop);
+hal_prop_t *hd_read_properties(const char *udi);
+int hd_write_properties(const char *udi, hal_prop_t *prop);
+
 /* implemented in hddb.c */
 
 str_list_t *get_hddb_packages(hd_data_t *hd_data);
@@ -2335,8 +2490,11 @@ hd_manual_t *hd_free_manual(hd_manual_t *manual);
 hd_t *hd_read_config(hd_data_t *hd_data, const char *id);
 int hd_write_config(hd_data_t *hd_data, hd_t *hd);
 char *hd_hw_item_name(hd_hw_item_t item);
+hd_hw_item_t hd_hw_item_type(char *name);
 char *hd_status_value_name(hd_status_value_t status);
 int hd_change_status(const char *id, hd_status_t status, const char *config_string);
+int hd_change_config_status(hd_data_t *hd_data, const char *id, hd_status_t status, const char *config_string);
+int hd_read_mmap(hd_data_t *hd_data, char *name, unsigned char *buf, off_t start, unsigned size);
 
 
 /*
