@@ -2,6 +2,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
+#include <fnmatch.h>
 #include <sys/utsname.h>
 
 #include "hd.h"
@@ -54,13 +55,15 @@ typedef struct {
   char *serial;
   str_list_t *driver;
   char *requires;
+  unsigned hwclass;
 } hddb_search_t;
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 static void hddb_init_pci(hd_data_t *hd_data);
-static hddb_pci_t *parse_pcimap(str_list_t *file);
-static driver_info_t *hd_pcidb(hd_data_t *hd_data, hddb_pci_t *pci_db, hd_t *hd);
+static char *get_mi_field(char *str, char *tag, int field_len, unsigned *value, unsigned *has_value);
+static modinfo_t *parse_modinfo(str_list_t *file);
+static driver_info_t *hd_modinfo_db(hd_data_t *hd_data, modinfo_t *modinfo_db, hd_t *hd, driver_info_t *drv_info);
 static void hddb_init_external(hd_data_t *hd_data);
 
 static line_t *parse_line(char *str);
@@ -98,114 +101,375 @@ static char *module_cmd(hd_t *hd, char *cmd);
 void hddb_init_pci(hd_data_t *hd_data)
 {
   str_list_t *sl = NULL;
-  char *s = NULL;
+  char *s = NULL, *r;
   struct utsname ubuf;
 
-  if(!hd_data->hddb_pci) {
+  if(!hd_data->modinfo) {
     if(!uname(&ubuf)) {
-      str_printf(&s, 0, "/lib/modules/%s/modules.pcimap", ubuf.release);
+      r = getenv("LIBHD_KERNELVERSION");
+      if(!r || !*r) r = ubuf.release;
+      str_printf(&s, 0, "/lib/modules/%s/modules.alias", r);
       sl = read_file(s, 0, 0);
       s = free_mem(s);
     }
 
-    hd_data->hddb_pci = parse_pcimap(sl);
+    hd_data->modinfo = parse_modinfo(sl);
 
     sl = free_str_list(sl);
   }
 
-  if(!hd_data->hddb_pci_hm) {
-    sl = read_file("/etc/hotplug/pci.handmap", 0, 0);
-    hd_data->hddb_pci_hm = parse_pcimap(sl);
-    sl = free_str_list(sl);
-  }
-}
-
-
-hddb_pci_t *parse_pcimap(str_list_t *file)
-{
-  str_list_t *sl;
-  unsigned len;
-  hddb_pci_t *pci, *p;
-  char buf[64];
-  unsigned u0, u1, u2, u3, u4, u5;
-
-  for(len = 1, sl = file; sl; sl = sl->next) len++;
-
-  pci = new_mem(len * sizeof *pci);
-
-  for(p = pci, sl = file; sl; sl = sl->next) {
-    if(sscanf(sl->str, "%63s %x %x %x %x %x %x", buf, &u0, &u1, &u2, &u3, &u4, &u5) == 7) {
-      p->module = new_str(buf);
-      p->vendor = u0;
-      p->device = u1;
-      p->subvendor = u2;
-      p->subdevice = u3;
-      p->pciclass = u4;
-      p->classmask = u5;
-
-      p++;
-    }
-  }
-
 #if 0
-  fprintf(stderr, "---  pcimap  ---\n");
-  for(p = pci; p->module; p++) {
-    fprintf(stderr, "%s, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x\n",
-      p->module, p->vendor, p->device, p->subvendor, p->subdevice,
-      p->pciclass, p->classmask
-    );
+  // currently nothing
+  if(!hd_data->modinfo_ext) {
+    sl = read_file("/WHATEVER", 0, 0);
+    hd_data->modinfo_ext = parse_modinfo(sl);
+    sl = free_str_list(sl);
   }
 #endif
-
-  return pci;
 }
 
 
-driver_info_t *hd_pcidb(hd_data_t *hd_data, hddb_pci_t *pci_db, hd_t *hd)
+char *get_mi_field(char *str, char *tag, int field_len, unsigned *value, unsigned *has_value)
 {
-  unsigned vendor, device, subvendor, subdevice, pciclass;
-  driver_info_t *di = NULL, *di0 = NULL;
-  pci_t *pci;
+  int i, j;
 
-  if(!pci_db) return NULL;
+  i = strlen(tag);
 
-  if(ID_TAG(hd->vendor.id) != TAG_PCI) return NULL;
+  if(strncmp(str, tag, i)) return NULL;
+  str += i;
 
-  vendor = ID_VALUE(hd->vendor.id);
-  device = ID_VALUE(hd->device.id);
-  subvendor = ID_VALUE(hd->sub_vendor.id);
-  subdevice = ID_VALUE(hd->sub_device.id);
-  pciclass = (hd->base_class.id << 16) + ((hd->sub_class.id & 0xff) << 8) + (hd->prog_if.id & 0xff);
-
-  if(
-    hd->detail &&
-    hd->detail->type == hd_detail_pci &&
-    (pci = hd->detail->pci.data)
-  ) {
-    pciclass = (pci->base_class << 16) + ((pci->sub_class & 0xff) << 8) + (pci->prog_if & 0xff);
+  if(*str == '*') {
+    str++;
+    *value = 0;
+    *has_value = 0;
+  }
+  else {
+    i = sscanf(str, "%8x%n", value, &j);
+    if(i < 1) return NULL;
+    *has_value = 1;
+    str += j;
   }
 
-  for(; pci_db->module; pci_db++) {
+  return str;
+}
+
+
+modinfo_t *parse_modinfo(str_list_t *file)
+{
+  str_list_t *sl;
+  unsigned len;
+  modinfo_t *modinfo, *m;
+  char *s;
+  unsigned u;
+  char alias[256], module[256];
+
+  /* lenght + 1! */
+  for(len = 1, sl = file; sl; sl = sl->next) len++;
+
+  modinfo = new_mem(len * sizeof *modinfo);
+
+  for(m = modinfo, sl = file; sl; sl = sl->next) {
+    if(sscanf(sl->str, "alias %255s %255s", alias, module) != 2) continue;
+
+    if(!strncmp(alias, "pci:", sizeof "pci:" - 1)) {
+      s = alias + sizeof "pci:" - 1;
+
+      if(!(s = get_mi_field(s, "v", 8, &m->pci.vendor, &u))) continue;
+      m->pci.has.vendor = u;
+
+      if(!(s = get_mi_field(s, "d", 8, &m->pci.device, &u))) continue;
+      m->pci.has.device = u;
+
+      if(!(s = get_mi_field(s, "sv", 8, &m->pci.sub_vendor, &u))) continue;
+      m->pci.has.sub_vendor = u;
+
+      if(!(s = get_mi_field(s, "sd", 8, &m->pci.sub_device, &u))) continue;
+      m->pci.has.sub_device = u;
+
+      if(!(s = get_mi_field(s, "bc", 2, &m->pci.base_class, &u))) continue;
+      m->pci.has.base_class = u;
+
+      if(!(s = get_mi_field(s, "sc", 2, &m->pci.sub_class, &u))) continue;
+      m->pci.has.sub_class = u;
+
+      if(!(s = get_mi_field(s, "i", 2, &m->pci.prog_if, &u))) continue;
+      m->pci.has.prog_if = u;
+
+      m->module = new_str(module);
+      m->alias = new_str(alias);
+      m++->type = mi_pci;
+
+      continue;
+    }
+
+    if(!strncmp(alias, "usb:", sizeof "usb:" - 1)) {
+      s = alias + sizeof "usb:" - 1;
+
+      m->module = new_str(module);
+      m->alias = new_str(alias);
+      m++->type = mi_usb;
+
+      continue;
+    }
+
+    if(!strncmp(alias, "pcmcia:", sizeof "pcmcia:" - 1)) {
+      s = alias + sizeof "pcmcia:" - 1;
+
+      m->module = new_str(module);
+      m->alias = new_str(alias);
+      m++->type = mi_pcmcia;
+
+      continue;
+    }
+
+  }
+
+  /* note: list stops at first entry with m->type == mi_none */
+
+#if 0
+  fprintf(stderr, "---  modinfo  ---\n");
+  for(m = modinfo; m->type; m++) {
+    switch(m->type) {
+      case mi_pci:
+        fprintf(stderr, "%s:%d, %s\n  v 0x%x:%u, d 0x%x:%u, sv 0x%x:%u, sd 0x%x:%u, bc 0x%x:%u, sc 0x%x:%u, i 0x%x:%u\n",
+          m->module, m->type, m->alias,
+          m->pci.vendor, m->pci.has.vendor,
+          m->pci.device, m->pci.has.device,
+          m->pci.sub_vendor, m->pci.has.sub_vendor,
+          m->pci.sub_device, m->pci.has.sub_device,
+          m->pci.base_class, m->pci.has.base_class,
+          m->pci.sub_class, m->pci.has.sub_class,
+          m->pci.prog_if, m->pci.has.prog_if
+        );
+        break;
+
+      case mi_usb:
+      case mi_pcmcia:
+        fprintf(stderr, "%s:%d, %s\n",
+          m->module, m->type, m->alias
+        );
+        break;
+
+      case mi_none:
+        break;
+    }
+  }
+#endif
+
+  return modinfo;
+}
+
+
+/* return prio, 0: no match */
+int match_modinfo(modinfo_t *db, modinfo_t *match)
+{
+  int prio = 0;
+  char *s;
+
+  if(db->type != match->type) return prio;
+
+  switch(db->type) {
+    case mi_pci:
+      if(db->pci.has.base_class) {
+        if(match->pci.has.base_class && db->pci.base_class == match->pci.base_class) {
+          prio = 1;
+        }
+        else {
+          prio = 0;
+          break;
+        }
+      }
+      if(db->pci.has.sub_class) {
+        if(match->pci.has.sub_class && db->pci.sub_class == match->pci.sub_class) {
+          prio = 1;
+        }
+        else {
+          prio = 0;
+          break;
+        }
+      }
+      if(db->pci.has.prog_if) {
+        if(match->pci.has.prog_if && db->pci.prog_if == match->pci.prog_if) {
+          prio = 1;
+        }
+        else {
+          prio = 0;
+          break;
+        }
+      }
+      if(db->pci.has.vendor) {
+        if(match->pci.has.vendor && db->pci.vendor == match->pci.vendor) {
+          prio = 2;
+        }
+        else {
+          prio = 0;
+          break;
+        }
+      }
+      if(db->pci.has.device) {
+        if(match->pci.has.device && db->pci.device == match->pci.device) {
+          prio = 3;
+        }
+        else {
+          prio = 0;
+          break;
+        }
+      }
+      if(db->pci.has.sub_vendor) {
+        if(match->pci.has.sub_vendor && db->pci.sub_vendor == match->pci.sub_vendor) {
+          prio = 4;
+        }
+        else {
+          prio = 0;
+          break;
+        }
+      }
+      if(db->pci.has.sub_device) {
+        if(match->pci.has.sub_device && db->pci.sub_device == match->pci.sub_device) {
+          prio = 5;
+        }
+        else {
+          prio = 0;
+          break;
+        }
+      }
+      break;
+
+    case mi_usb:
+    case mi_pcmcia:
+      if(match->alias && db->alias) {
+        if(!fnmatch(db->alias, match->alias, 0)) {
+          s = strchr(db->alias, '*');
+          prio = s ? s - db->alias + 1 : strlen(db->alias) + 1;
+        }
+      }
+      break;
+
+    case mi_none:
+      return 0;
+  }
+
+  return prio;
+}
+
+
+driver_info_t *hd_modinfo_db(hd_data_t *hd_data, modinfo_t *modinfo_db, hd_t *hd, driver_info_t *drv_info)
+{
+  driver_info_t **di = NULL, *di2;
+  pci_t *pci;
+  char *mod_list[16 /* arbitrary */];
+  int mod_prio[sizeof mod_list / sizeof *mod_list];
+  int i, prio, mod_list_len;
+  modinfo_t match = { };
+
+  if(!modinfo_db) return drv_info;
+
+  if(
+    ID_TAG(hd->vendor.id) == TAG_PCI ||
+    ID_TAG(hd->device.id) == TAG_PCI
+  ) {
+    match.type = mi_pci;
+  }
+  else if(
+    ID_TAG(hd->vendor.id) == TAG_USB ||
+    ID_TAG(hd->device.id) == TAG_USB
+  ) {
+    match.type = mi_usb;
+  }
+  else if(
+    ID_TAG(hd->vendor.id) == TAG_PCMCIA ||
+    ID_TAG(hd->device.id) == TAG_PCMCIA ||
+    (hd->modalias && !strncmp(hd->modalias, "pcmcia:", sizeof "pcmcia:" - 1))
+  ) {
+    match.type = mi_pcmcia;
+  }
+
+  match.alias = hd->modalias;
+  
+  if(!match.type) return drv_info;
+
+  /* don't add module info if driver info of some other type exists */
+  for(di = &drv_info; *di; di = &(*di)->next) {
+    if((*di)->any.type != di_module) return drv_info;
+  }
+
+  if(match.type == mi_pci) {
+    if(hd->vendor.id) {
+      match.pci.vendor = ID_VALUE(hd->vendor.id);
+      match.pci.has.vendor = 1;
+    }
+    if(hd->device.id) {
+      match.pci.device = ID_VALUE(hd->device.id);
+      match.pci.has.device = 1;
+    }
+    if(hd->sub_vendor.id) {
+      match.pci.sub_vendor = ID_VALUE(hd->sub_vendor.id);
+      match.pci.has.sub_vendor = 1;
+    }
+    if(hd->sub_device.id) {
+      match.pci.sub_device = ID_VALUE(hd->sub_device.id);
+      match.pci.has.sub_device = 1;
+    }
+    match.pci.base_class = hd->base_class.id;
+    match.pci.has.base_class = 1;
+    match.pci.sub_class = hd->sub_class.id;
+    match.pci.has.sub_class = 1;
+    match.pci.prog_if = hd->prog_if.id;
+    match.pci.has.prog_if = 1;
+
     if(
-      (pci_db->vendor == 0xffffffff || pci_db->vendor == vendor) &&
-      (pci_db->device == 0xffffffff || pci_db->device == device) &&
-      (pci_db->subvendor == 0xffffffff || pci_db->subvendor == subvendor) &&
-      (pci_db->subdevice == 0xffffffff || pci_db->subdevice == subdevice) &&
-      !((pci_db->pciclass ^ pciclass) & pci_db->classmask)
+      hd->detail &&
+      hd->detail->type == hd_detail_pci &&
+      (pci = hd->detail->pci.data)
     ) {
-      if(di) {
-        di = di->next = new_mem(sizeof *di);
-      }
-      else {
-        di = di0 = new_mem(sizeof *di);
-      }
-      di->any.type = di_module;
-      di->module.modprobe = 1;
-      add_str_list(&di->any.hddb0, pci_db->module);
+      match.pci.base_class = pci->base_class;
+      match.pci.sub_class = pci->sub_class;
+      match.pci.prog_if = pci->prog_if;
     }
   }
 
-  return di0;
+  for(mod_list_len = 0; modinfo_db->type; modinfo_db++) {
+    if((prio = match_modinfo(modinfo_db, &match))) {
+      for(di2 = drv_info; di2; di2 = di2->next) {
+        if(
+          di2->any.type == di_module &&
+          di2->any.hddb0 &&
+          di2->any.hddb0->str &&
+          !strcmp(di2->any.hddb0->str, modinfo_db->module)
+        ) break;
+      }
+
+      if(di2) continue;
+
+      for(i = 0; i < mod_list_len; i++) {
+        if(!strcmp(mod_list[i], modinfo_db->module)) {
+          if(prio > mod_prio[i]) mod_prio[i] = prio;
+          break;
+        }
+      }
+
+      if(i < mod_list_len) continue;
+
+      mod_prio[mod_list_len] = prio;
+      mod_list[mod_list_len++] = modinfo_db->module;
+
+      if(mod_list_len >= sizeof mod_list / sizeof *mod_list) break;
+    }
+  }
+
+  for(prio = 256; prio >= 0; prio--) {
+    for(i = 0; i < mod_list_len; i++) {
+      if(mod_prio[i] == prio) {
+        *di = new_mem(sizeof **di);
+        (*di)->any.type = di_module;
+        (*di)->module.modprobe = 1;
+        add_str_list(&(*di)->any.hddb0, mod_list[i]);
+        di = &(*di)->next;
+      }
+    }
+  }
+
+  return drv_info;
 }
 
 
@@ -241,7 +505,7 @@ void hddb_init_external(hd_data_t *hd_data)
 
   hddb2 = hd_data->hddb2[0] = new_mem(sizeof *hd_data->hddb2[0]);
 
-  sl0 = read_file(ID_LIST, 0, 0);
+  sl0 = read_file(hd_get_hddb_path("hd.ids"), 0, 0);
 
   l_start = l_end = 0;
   state = 0;
@@ -570,6 +834,7 @@ int parse_id(char *str, unsigned *id, unsigned *range, unsigned *mask)
     else if(!strcmp(s, "special")) tag = TAG_SPECIAL;
     else if(!strcmp(s, "eisa")) tag = TAG_EISA;
     else if(!strcmp(s, "isapnp")) tag = TAG_EISA;
+    else if(!strcmp(s, "pcmcia")) tag = TAG_PCMCIA;
     else {
       str = s;
       if(t) *t = c;	/* restore */
@@ -617,6 +882,7 @@ hddb_entry_mask_t add_entry(hddb2_data_t *hddb2, tmp_entry_t *te, hddb_entry_t i
   int i;
   unsigned u, u0, u1, u2;
   char *s, c;
+  str_list_t *sl, *sl0;
 
   for(i = 0; (unsigned) i < sizeof hddb_is_numeric / sizeof *hddb_is_numeric; i++) {
     if(idx == hddb_is_numeric[i]) break;
@@ -626,7 +892,23 @@ hddb_entry_mask_t add_entry(hddb2_data_t *hddb2, tmp_entry_t *te, hddb_entry_t i
     /* numeric id */
     mask |= 1 << idx;
 
-    i = parse_id(str, &u0, &u1, &u2);
+    /* special */
+    if(idx == he_hwclass) {
+      sl0 = hd_split('|', str);
+      for(u0 = u1 = 0, sl = sl0; sl && u1 <= 16; sl = sl->next) {
+        u = hd_hw_item_type(sl->str);
+        if(u) {
+          u0 += u << u1;
+          u1 += 8;
+        }
+      }
+      free_str_list(sl0);
+
+      i = 1;
+    }
+    else {
+      i = parse_id(str, &u0, &u1, &u2);
+    }
 
     switch(i) {
       case 1:
@@ -804,7 +1086,7 @@ void hddb_dump_skey(hddb2_data_t *hddb, FILE *f, prefix_t pre, hddb_entry_mask_t
   hddb_entry_t ent;
   unsigned rm_val = 0, r_or_m = 0;
   unsigned fl, val, *ids, id, tag, u;
-  char *str_val;
+  char *str_val, *s;
   int i;
 
   if(pre >= sizeof pref_char) return;
@@ -848,7 +1130,15 @@ void hddb_dump_skey(hddb2_data_t *hddb, FILE *f, prefix_t pre, hddb_entry_mask_t
       if(fl == FLAG_ID) {
         tag = ID_TAG(val);
         id = ID_VALUE(val);
-        if(tag == TAG_EISA && (ent == he_vendor_id || ent == he_subvendor_id)) {
+        if(ent == he_hwclass) {
+          /* is special */
+          for(u = (val & 0xffffff); u; u >>= 8) {
+            s = hd_hw_item_name(u & 0xff);
+            if(s) fprintf(f, "%s", s);
+            if(u > 0x100) fprintf(f, "|");
+          }
+        }
+        else if(tag == TAG_EISA && (ent == he_vendor_id || ent == he_subvendor_id)) {
           fprintf(f, "%s", eisa_vendor_str(id));
         }
         else {
@@ -1038,10 +1328,16 @@ int compare_ids(hddb2_data_t *hddb, hddb_search_t *hs, hddb_entry_mask_t mask, u
         case he_rev_id:
           id = hs->revision.id;
           break;
+
 	case he_detail_ccw_data_cu_model:
 	  id = hs->cu_model.id;
 	  break;
 
+#if 0
+	/* not allowed as search key */
+	case he_hwclass:
+#endif
+
         default:
           ok = 0;
           break;
@@ -1170,6 +1466,7 @@ int compare_ids(hddb2_data_t *hddb, hddb_search_t *hs, hddb_entry_mask_t mask, u
   return res;
 }
 
+
 void complete_ids(
   hddb2_data_t *hddb, hddb_search_t *hs,
   hddb_entry_mask_t key_mask, hddb_entry_mask_t mask, unsigned val_idx
@@ -1234,11 +1531,15 @@ void complete_ids(
         case he_rev_id:
           id = &hs->revision.id;
           break;
-	
+
 	case he_detail_ccw_data_cu_model:
 	  id = &hs->cu_model.id;
 	  break;
 
+	case he_hwclass:
+	  id = &hs->hwclass;
+	  break;
+
         default:
           ok = 0;
           break;
@@ -1497,6 +1798,7 @@ void hddb_add_info(hd_data_t *hd_data, hd_t *hd)
 {
   hddb_search_t hs = {};
   driver_info_t *new_driver_info = NULL;
+  unsigned u;
 #if WITH_ISDN
   cdb_isdn_card *cic;
 #endif
@@ -1647,6 +1949,12 @@ void hddb_add_info(hd_data_t *hd_data, hd_t *hd)
       hd->detail->ccw.data->cu_model=hs.cu_model.id;
   }
 
+  if((hs.value & (1 << he_hwclass))) {
+    for(u = hs.hwclass; u; u >>= 8) {
+      hd_set_hw_class(hd, u & 0xff);
+    }
+  }
+
   /* look for sub vendor again */
 
   if(!hd->sub_vendor.name && hd->sub_vendor.id) {
@@ -1688,6 +1996,24 @@ void hddb_add_info(hd_data_t *hd_data, hd_t *hd)
     }
   }
 
+  /* get package info for compat device id */
+
+  if(!hd->requires) {
+    hddb_search_t hs2 = {};
+
+    hs2.vendor.id = hd->compat_vendor.id;
+    hs2.key |= 1 << he_vendor_id;
+
+    hs2.device.id = hd->compat_device.id;
+    hs2.key |= 1 << he_device_id;
+
+    hddb_search(hd_data, &hs2, 1);
+
+    if((hs2.value & (1 << he_requires))) {
+      hd->requires = hd_split('|', hs2.requires);
+    }
+  }
+
   /* get driver info */
 
 #if WITH_ISDN
@@ -1708,9 +2034,10 @@ void hddb_add_info(hd_data_t *hd_data, hd_t *hd)
 #endif
 
   if(!new_driver_info) {
-    new_driver_info = hd_pcidb(hd_data, hd_data->hddb_pci_hm, hd);
+    new_driver_info = hd_modinfo_db(hd_data, hd_data->modinfo_ext, hd, new_driver_info);
   }
 
+#if 1
   if(!new_driver_info && (hs.value & (1 << he_driver))) {
     new_driver_info = hddb_to_device_driver(hd_data, &hs);
   }
@@ -1733,6 +2060,24 @@ void hddb_add_info(hd_data_t *hd_data, hd_t *hd)
       new_driver_info =  hddb_to_device_driver(hd_data, &hs);
     }
   }
+#endif
+
+  /* acpi: load temperature control modules */
+  if(!new_driver_info && hd->is.with_acpi) {
+    memset(&hs, 0, sizeof hs);
+
+    hs.vendor.id = MAKE_ID(TAG_SPECIAL, 0xf001);
+    hs.key |= 1 << he_vendor_id;
+
+    hs.device.id = MAKE_ID(TAG_SPECIAL, 4);
+    hs.key |= 1 << he_device_id;
+
+    hddb_search(hd_data, &hs, 1);
+
+    if((hs.value & (1 << he_driver))) {
+      new_driver_info =  hddb_to_device_driver(hd_data, &hs);
+    }
+  }
 
   if(!new_driver_info && hd->base_class.id == bc_keyboard) {
     new_driver_info = kbd_driver(hd_data, hd);
@@ -1742,9 +2087,7 @@ void hddb_add_info(hd_data_t *hd_data, hd_t *hd)
     new_driver_info = monitor_driver(hd_data, hd);
   }
 
-  if(!new_driver_info) {
-    new_driver_info = hd_pcidb(hd_data, hd_data->hddb_pci, hd);
-  }
+  new_driver_info = hd_modinfo_db(hd_data, hd_data->modinfo, hd, new_driver_info);
 
   if(new_driver_info) {
     if(!hd->ref) {
@@ -2481,7 +2824,10 @@ void expand_driver_info(hd_data_t *hd_data, hd_t *hd)
           t = strsep(&t0, " ");
 
           add_str_list(&di->module.names, t);
-          di->module.active &= hd_module_is_active(hd_data, t);
+          di->module.active &= (
+            hd_module_is_active(hd_data, t) |
+            (search_str_list(hd->drivers, t) ? 1 : 0)
+          );
 
           if(t0) {
             add_str_list(&di->module.mod_args, module_cmd(hd, t0));
@@ -2500,12 +2846,22 @@ void expand_driver_info(hd_data_t *hd_data, hd_t *hd)
 
       case di_mouse:
         di->mouse.buttons = di->mouse.wheels = -1;
+        u1 = 0;
+        if(
+          hd->compat_vendor.id == MAKE_ID(TAG_SPECIAL, 0x0210) &&
+          ID_TAG(hd->compat_device.id) == TAG_SPECIAL
+        ) {
+          u1 = hd->compat_device.id;
+        }
         if(
           hd->vendor.id == MAKE_ID(TAG_SPECIAL, 0x0210) &&
           ID_TAG(hd->device.id) == TAG_SPECIAL
         ) {
-          di->mouse.wheels = ID_VALUE(hd->device.id) >> 4;
-          di->mouse.buttons = ID_VALUE(hd->device.id) & 15;
+          u1 = hd->device.id;
+        }
+        if(u1) {
+          di->mouse.wheels = ID_VALUE(u1) >> 4;
+          di->mouse.buttons = ID_VALUE(u1) & 15;
         }
         for(i = 0, sl = di->mouse.hddb0; sl; sl = sl->next, i++) {
           if(i == 0) {
@@ -2616,7 +2972,7 @@ void expand_driver_info(hd_data_t *hd_data, hd_t *hd)
   if(di && di->any.type == di_x11 && !hd_probe_feature(hd_data, pr_ignx11)) {
     cmd = get_cmdline(hd_data, "x11");
     if(cmd && *cmd->str) {
-      di = reorder_x11(di, cmd->str);
+      hd->driver_info = reorder_x11(di, cmd->str);
     }
     free_str_list(cmd);
   }
