@@ -46,6 +46,7 @@ void hd_dump_entry(hd_data_t *hd_data, hd_t *h, FILE *f)
   hd_t *hd_tmp;
   int i, j;
   str_list_t *sl;
+  hal_prop_t *prop;
 
 #ifdef LIBHD_MEMCHECK
   {
@@ -97,6 +98,14 @@ void hd_dump_entry(hd_data_t *hd_data, hd_t *h, FILE *f)
 
   a2 = free_mem(a2);
 
+  if(h->udi) {
+    dump_line("UDI: %s\n", h->udi);
+  }
+
+  if(h->parent_udi) {
+    dump_line("Parent UDI: %s\n", h->parent_udi);
+  }
+
   if((hd_data->debug & HD_DEB_CREATION) && h->unique_id) {
     dump_line("Unique ID: %s\n", h->unique_id);
   }
@@ -146,16 +155,18 @@ void hd_dump_entry(hd_data_t *hd_data, hd_t *h, FILE *f)
     if(j) dump_line0("\n");
   }
 
-  if(h->base_class.id == bc_internal && h->sub_class.id == sc_int_cpu)
+  if(h->base_class.id == bc_internal && h->sub_class.id == sc_int_cpu) {
     dump_cpu(hd_data, h, f);
-  else if(h->base_class.id == bc_internal && h->sub_class.id == sc_int_bios)
+  }
+  else if(h->base_class.id == bc_internal && h->sub_class.id == sc_int_bios) {
     dump_bios(hd_data, h, f);
-  else if(h->base_class.id == bc_internal && h->sub_class.id == sc_int_prom)
+  }
+  else if(h->base_class.id == bc_internal && h->sub_class.id == sc_int_prom) {
     dump_prom(hd_data, h, f);
-  else if(h->base_class.id == bc_internal && h->sub_class.id == sc_int_sys)
-    dump_sys(hd_data, h, f);
-  else
+  }
+  else {
     dump_normal(hd_data, h, f);
+  }
 
   s1 = s2 = NULL;
   if(h->is.notready) {
@@ -182,6 +193,22 @@ void hd_dump_entry(hd_data_t *hd_data, hd_t *h, FILE *f)
     dump_line0("\n");
   }
 
+  if(hd_data->debug == -1 && (prop = h->hal_prop)) {
+    dump_line_str("HAL Properties:\n");
+    for(; prop; prop = prop->next) {
+      s = hd_hal_print_prop(prop);
+      dump_line("  %s\n", s);
+    }
+  }
+
+  if(hd_data->debug == -1 && (prop = h->persistent_prop)) {
+    dump_line_str("Persistent Properties:\n");
+    for(; prop; prop = prop->next) {
+      s = hd_hal_print_prop(prop);
+      dump_line("  %s\n", s);
+    }
+  }
+
   if(
     hd_data->debug && (
       h->status.configured ||
@@ -242,9 +269,9 @@ void hd_dump_entry(hd_data_t *hd_data, hd_t *h, FILE *f)
 
   if(h->detail && h->detail->ccw.type==hd_detail_ccw)
   {
-  	  dump_line("LCSS: %x.%x\n",h->detail->ccw.data->lcss >> 8, h->detail->ccw.data->lcss & 0xf);
-	  dump_line("CU Model: 0x%x\n",h->detail->ccw.data->cu_model);
-	  dump_line("Device Model: 0x%x\n",h->detail->ccw.data->dev_model);
+    dump_line("LCSS: %x.%x\n",h->detail->ccw.data->lcss >> 8, h->detail->ccw.data->lcss & 0xf);
+    dump_line("CU Model: 0x%x\n",h->detail->ccw.data->cu_model);
+    dump_line("Device Model: 0x%x\n",h->detail->ccw.data->dev_model);
   }
 
 #if defined(__s390__) || defined(__s390x__)
@@ -262,7 +289,8 @@ void hd_dump_entry(hd_data_t *hd_data, hd_t *h, FILE *f)
     h->base_class.id == bc_storage_device &&
     h->sub_class.id == sc_sdev_cdrom &&
     h->detail &&
-    h->detail->type == hd_detail_cdrom
+    h->detail->type == hd_detail_cdrom &&
+    h->detail->cdrom.data
   ) {
     cdrom_info_t *ci = h->detail->cdrom.data;
 
@@ -357,6 +385,7 @@ void dump_normal(hd_data_t *hd_data, hd_t *h, FILE *f)
   driver_info_t *di;
   str_list_t *sl, *sl1, *sl2;
   isdn_parm_t *ip;
+  static char *geo_type_str[] = { "Physical", "Logical", "BIOS EDD", "BIOS Legacy" };
 
   if(h->model) dump_line("Model: \"%s\"\n", h->model);
 
@@ -428,18 +457,16 @@ void dump_normal(hd_data_t *hd_data, hd_t *h, FILE *f)
     );
   }
 
+  if(h->base_class.id == bc_internal && h->sub_class.id == sc_int_sys) {
+    dump_sys(hd_data, h, f);
+  }
+
   if(h->drivers) {
     s = hd_join("\", \"", h->drivers);
     dump_line("Driver: \"%s\"\n", s);
     s = free_mem(s);
   }
 
-#if 0
-  if(h->driver) {
-    dump_line("Driver: \"%s\"\n", h->driver);
-  }
-#endif
-
   if(h->broken) {
     dump_line_str("Warning: might be broken\n");
   }
@@ -488,7 +515,7 @@ void dump_normal(hd_data_t *hd_data, hd_t *h, FILE *f)
     dump_line("PROM id: %s\n", h->rom_id);
 #endif
 #if defined(__s390__) || defined(__s390x__)
-    dump_line("Chp id: %s\n", h->rom_id);
+    dump_line("IUCV user: %s\n", h->rom_id);
 #endif
   }
 
@@ -618,6 +645,9 @@ void dump_normal(hd_data_t *hd_data, hd_t *h, FILE *f)
                 case size_unit_cm:
                   s = "cm";
                   break;
+                case size_unit_mm:
+                  s = "mm";
+                  break;
                 case size_unit_sectors:
                   s = "sectors";
                   break;
@@ -640,11 +670,15 @@ void dump_normal(hd_data_t *hd_data, hd_t *h, FILE *f)
         break;
 
       case res_disk_geo:
-        s = res->disk_geo.logical ? "Logical" : "Physical";
+        s = res->disk_geo.geotype < sizeof geo_type_str / sizeof *geo_type_str ?
+          geo_type_str[res->disk_geo.geotype] : "";
         dump_line(
           "Geometry (%s): CHS %u/%u/%u\n", s,
           res->disk_geo.cyls, res->disk_geo.heads, res->disk_geo.sectors
         );
+        if(res->disk_geo.size) {
+          dump_line("Size (%s): %"PRIu64" sectors\n", s, res->disk_geo.size);
+        }
         break;
 
       case res_cache:
@@ -705,8 +739,46 @@ void dump_normal(hd_data_t *hd_data, hd_t *h, FILE *f)
 	dump_line("HW Address: %s\n", res->hwaddr.addr);
         break;
 
+      case res_link:
+	dump_line("Link detected: %s\n", res->link.state ? "yes" : "no");
+        break;
+
+      case res_wlan:
+	if(res->wlan.channels) {
+		str_list_t *ptr = res->wlan.channels;
+		dump_line("WLAN channels: %s", ptr->str);
+		while((ptr=ptr->next)) dump_line0(" %s", ptr->str);
+		dump_line0("\n");
+	}
+	if(res->wlan.frequencies) {
+		str_list_t *ptr = res->wlan.frequencies;
+		dump_line("WLAN frequencies: %s", ptr->str);
+		while((ptr=ptr->next)) dump_line0(" %s", ptr->str);
+		dump_line0("\n");
+	}
+	if(res->wlan.bitrates) {
+		str_list_t *ptr = res->wlan.bitrates;
+		dump_line("WLAN bitrates: %s", ptr->str);
+		while((ptr=ptr->next)) dump_line0(" %s", ptr->str);
+		dump_line0("\n");
+	}
+	if(res->wlan.enc_modes) {
+		str_list_t *ptr = res->wlan.enc_modes;
+		dump_line("WLAN encryption modes: %s", ptr->str);
+		while((ptr=ptr->next)) dump_line0(" %s", ptr->str);
+		dump_line0("\n");
+	}
+        if(res->wlan.auth_modes) {
+		str_list_t *ptr = res->wlan.auth_modes;
+		dump_line("WLAN authentication modes: %s", ptr->str);
+		while((ptr=ptr->next)) dump_line0(" %s", ptr->str);
+		dump_line0("\n");
+	}
+        break;
+
+
       default:
-        dump_line("Unkown resource type %d\n", res->any.type);
+        dump_line("Unknown resource type %d\n", res->any.type);
     }
   }
 
@@ -718,6 +790,10 @@ void dump_normal(hd_data_t *hd_data, hd_t *h, FILE *f)
     dump_line0("\n");
   }
 
+  if(h->modalias) {
+    dump_line("Module Alias: \"%s\"\n", h->modalias);
+  }
+
   for(di = h->driver_info, i = 0; di; di = di->next, i++) {
     dump_line("Driver Info #%d:\n", i);
     ind += 2;
@@ -1035,7 +1111,7 @@ void dump_bios(hd_data_t *hd_data, hd_t *hd, FILE *f)
   if(bt->par_port1) dump_line("Parallel Port 1: 0x%x\n", bt->par_port1);
   if(bt->par_port2) dump_line("Parallel Port 2: 0x%x\n", bt->par_port2);
 
-  if(bt->low_mem_size) dump_line("Base Memory: %uk\n", bt->low_mem_size >> 10);
+  if(bt->low_mem_size) dump_line("Base Memory: %u kB\n", bt->low_mem_size >> 10);
 
   if(bt->is_pnp_bios) {
     char *s = isa_id2str(bt->pnp_id);
@@ -1102,18 +1178,15 @@ void dump_sys(hd_data_t *hd_data, hd_t *hd, FILE *f)
   if(st->system_type) {
     dump_line("SystemType: \"%s\"\n", st->system_type);
   }
+
   if(st->generation) {
     dump_line("Generation: \"%s\"\n", st->generation);
   }
-  if(st->vendor) {
-    dump_line("Vendor: \"%s\"\n", st->vendor);
-  }
-  if(st->model) {
-    dump_line("Model: \"%s\"\n", st->model);
-  }
-  if(st->serial) {
-    dump_line("Serial ID: \"%s\"\n", st->serial);
+
+  if(st->formfactor) {
+    dump_line("Formfactor: \"%s\"\n", st->formfactor);
   }
+
   if(st->lang) {
     dump_line("Language: \"%s\"\n", st->lang);
   }
