@@ -37,6 +37,12 @@ static void int_devicenames(hd_data_t *hd_data);
 #if defined(__i386__) || defined (__x86_64__)
 static void int_softraid(hd_data_t *hd_data);
 #endif
+#if defined(__i386__) || defined (__x86_64__)
+static void int_system(hd_data_t *hd_data);
+static void int_legacy_geo(hd_data_t *hd_data);
+#endif
+static void int_find_parent(hd_data_t *hd_data);
+
 
 void hd_scan_int(hd_data_t *hd_data)
 {
@@ -66,10 +72,17 @@ void hd_scan_int(hd_data_t *hd_data)
   PROGRESS(6, 0, "mouse");
   int_mouse(hd_data);
 
+#if defined(__i386__) || defined (__x86_64__)
+  PROGRESS(15, 0, "system info");
+  int_system(hd_data);
+#endif
+
   PROGRESS(7, 0, "hdb");
+  hd_data->flags.keep_kmods = 1;
   for(hd = hd_data->hd; hd; hd = hd->next) {
     hddb_add_info(hd_data, hd);
   }
+  hd_data->flags.keep_kmods = 0;
 
   PROGRESS(8, 0, "usbscsi");
   int_fix_usb_scsi(hd_data);
@@ -92,7 +105,13 @@ void hd_scan_int(hd_data_t *hd_data)
 #if defined(__i386__) || defined (__x86_64__)
   PROGRESS(14, 0, "soft raid");
   int_softraid(hd_data);
+
+  PROGRESS(15, 0, "geo");
+  int_legacy_geo(hd_data);
 #endif
+
+  PROGRESS(16, 0, "parent");
+  int_find_parent(hd_data);
 }
 
 /*
@@ -100,70 +119,13 @@ void hd_scan_int(hd_data_t *hd_data)
  */
 void int_hotplug(hd_data_t *hd_data)
 {
-  hd_t *hd, *hd1, *bridge_hd;
-  unsigned p_sock[8], p_socks, u = 0;
+  hd_t *hd;
 
   for(hd = hd_data->hd; hd; hd = hd->next) {
     if(hd->bus.id == bus_usb || hd->usb_guid) {
       hd->hotplug = hp_usb;
     }
-    if((bridge_hd = hd_get_device_by_idx(hd_data, hd->attached_to))) {
-      if(
-        bridge_hd->base_class.id == bc_bridge &&
-        bridge_hd->sub_class.id == sc_bridge_cardbus
-      ) {
-        hd->hotplug = hp_cardbus;
-      }
-     else if(
-        bridge_hd->base_class.id == bc_bridge &&
-        bridge_hd->sub_class.id == sc_bridge_pcmcia
-      ) {
-        hd->hotplug = hp_pcmcia;
-      }
-    }
   }
-
-  for(p_socks = 0, hd = hd_data->hd; hd; hd = hd->next) {
-    if(
-      u < sizeof p_sock / sizeof *p_sock &&
-      is_pcmcia_ctrl(hd_data, hd)
-    ) {
-      p_sock[p_socks++] = hd->idx;
-    }
-  }
-
-  if(p_socks) {
-    for(hd = hd_data->hd; hd; hd = hd->next) {
-      if(
-        !hd->tag.remove &&
-        hd->bus.id == bus_pcmcia &&
-        hd->slot < p_socks &&
-        p_sock[hd->slot]
-      ) {
-        for(u = p_sock[hd->slot], hd1 = hd_data->hd; hd1; hd1 = hd1->next) {
-          if(hd1->tag.remove) continue;
-          if(hd1->status.available == status_no) continue;
-          if(hd1->attached_to == u) break;
-        }
-        if(hd1) {
-          hd1->hotplug = hd->hotplug;
-          hd1->hotplug_slot = hd->hotplug_slot;
-          if(!hd1->extra_info) {
-            hd1->extra_info = hd->extra_info;
-            hd->extra_info = NULL;
-          }
-          hd->tag.remove = 1;
-        }
-        else {
-          hd->attached_to = p_sock[hd->slot];
-        }
-        p_sock[hd->slot] = 0;
-      }
-    }
-
-    remove_tagged_hd_entries(hd_data);
-  }
-
 }
 
 /*
@@ -600,7 +562,6 @@ void int_floppy(hd_data_t *hd_data)
 }
 
 
-#define COPY_ENTRY(a) if(hd_scsi->a) { free_mem(hd_usb->a); hd_usb->a = new_str(hd_scsi->a); }
 /*
  * Remove usb entries that are handled by usb-storage.
  */
@@ -630,6 +591,24 @@ void int_fix_usb_scsi(hd_data_t *hd_data)
             add_res_entry(&hd_scsi->res, hd_usb->res);
             hd_usb->res = NULL;
 
+            if(!hd_scsi->modalias) {
+              hd_scsi->modalias = hd_usb->modalias;
+              hd_usb->modalias = NULL;
+            }
+
+            if(!hd_scsi->vendor.id) hd_scsi->vendor.id = hd_usb->vendor.id;
+            if(!hd_scsi->device.id) hd_scsi->device.id = hd_usb->device.id;
+
+            if(!hd_scsi->serial) {
+              hd_scsi->serial = hd_usb->serial;
+              hd_usb->serial = NULL;
+            }
+
+            if(!hd_scsi->driver_info) {
+              hd_scsi->driver_info = hd_usb->driver_info;
+              hd_usb->driver_info = NULL;
+            }
+
             new_id(hd_data, hd_scsi);
 
             hd_usb->tag.remove = 1;
@@ -637,75 +616,10 @@ void int_fix_usb_scsi(hd_data_t *hd_data)
         }
       }
     }
-
   }
 
-
-
-#if 0
-  for(hd_scsi = hd_data->hd; hd_scsi; hd_scsi = hd_scsi->next) {
-    if(
-      hd_scsi->bus.id == bus_scsi &&
-      hd_scsi->usb_guid &&
-      search_str_list(hd_scsi->drivers, "usb-storage")
-    ) {
-      for(hd_usb = hd_data->hd; hd_usb ; hd_usb = hd_usb->next) {
-        if(
-          hd_usb->bus.id == bus_usb &&
-          hd_usb->usb_guid &&
-          !hd_usb->tag.remove &&
-          !strcmp(hd_usb->usb_guid, hd_scsi->usb_guid) &&
-          hd_usb->detail &&
-          hd_usb->detail->type == hd_detail_usb &&
-          (usb = hd_usb->detail->usb.data) &&
-          usb->driver &&
-          !strcmp(usb->driver, "usb-storage")
-        ) {
-          hd_scsi->tag.remove = 1;
-
-          /* join usb & scsi info */
-          hd_usb->bus.id = hd_scsi->bus.id;
-          COPY_ENTRY(bus.name);
-          hd_usb->base_class.id = hd_scsi->base_class.id;
-          COPY_ENTRY(base_class.name);
-          hd_usb->sub_class.id = hd_scsi->sub_class.id;
-          COPY_ENTRY(sub_class.name);
-          hd_usb->prog_if.id = hd_scsi->prog_if.id;
-          COPY_ENTRY(prog_if.name);
-          COPY_ENTRY(unix_dev_name);
-          COPY_ENTRY(model);
-          // ###### FIXME?: COPY_ENTRY(driver)
-
-          hd_usb->vendor.id = hd_scsi->vendor.id;
-          COPY_ENTRY(vendor.name);
-          hd_usb->device.id = hd_scsi->device.id;
-          COPY_ENTRY(device.name);
-          hd_usb->sub_vendor.name = free_mem(hd_usb->sub_vendor.name);
-          COPY_ENTRY(sub_vendor.name);
-          hd_usb->sub_device.name = free_mem(hd_usb->sub_device.name);
-          COPY_ENTRY(sub_device.name);
-          COPY_ENTRY(revision.name);
-          COPY_ENTRY(serial);
-
-          hd_usb->is.notready = hd_scsi->is.notready;
-          if(hd_usb->block0) free_mem(hd_usb->block0);
-          hd_usb->block0 = hd_scsi->block0;
-          hd_scsi->block0 = NULL;
-          add_res_entry(&hd_usb->res, hd_scsi->res);
-          hd_scsi->res = NULL;
-
-          new_id(hd_data, hd_usb);
-        }
-      }
-    }
-  }
-#endif
-
-
-
   remove_tagged_hd_entries(hd_data);
 }
-#undef COPY_ENTRY
 
 
 /*
@@ -786,7 +700,7 @@ void int_modem(hd_data_t *hd_data)
           dev_num.minor = 1;
           break;
         case sc_mod_win2:
-          s = new_str("/dev/536ep");
+          s = new_str("/dev/536ep0");
           dev_num.major = 240;
           dev_num.minor = 1;
           break;
@@ -824,18 +738,33 @@ void int_wlan(hd_data_t *hd_data)
   str_list_t *sl;
   unsigned u, found;
   static char *wlan_mods[] = {
+    "acx_pci",
     "airo",
     "airo_cs",
     "aironet4500_card",
     "aironet4500_cs",
     "airport",
+    "adm8211",
     "arlan",
+    "at76c503",
+    "at76c503-i3861",
+    "at76c503-i3863",
+    "at76c503-rfmd",
+    "at76c503-rfmd-acc",
+    "at76c505-rfmd",
+    "at76c505-rfmd2958",
     "ath_hal",
     "ath_pci",
+    "atmel",
+    "atmel_cs",
+    "atmel_pci",
     "hermes",
     "hostap",
     "hostap_pci",
     "hostap_plx",
+    "ipw2100",
+    "ipw2200",
+    "ipw3945",
     "netwave_cs",
     "orinoco_cs",
     "orinoco_pci",
@@ -846,24 +775,19 @@ void int_wlan(hd_data_t *hd_data)
     "prism2_plx",
     "prism2_usb",
     "prism54",
+    "r8180",
     "ray_cs",
+    "rt2400",
+    "rt2500",
+    "rt2570",
+    "rt2400pci",
+    "rt2500pci",
+    "rt2500usb",
+    "usbdfu",
     "wavelan",
     "wavelan_cs",
-    "ipw2100",
-    "acx100_pci",
-    "atmel",
-    "atmel_cs",
-    "atmel_pci",
-    "at76c503-i3861",
-    "at76c503-i3863",
-    "at76c503-rfmd-acc",
-    "at76c503-rfmd",
-    "at76c503",
-    "at76c505-rfmd",
-    "at76c505-rfmd2958",
-    "usbdfu",
     "wl3501_cs",
-    "ipw2100"
+    "zd1201"
   };
 
   for(hd = hd_data->hd; hd; hd = hd->next) {
@@ -975,12 +899,13 @@ void int_softraid(hd_data_t *hd_data)
   /* no disks -> no check necessary */
   if(!hd) return;
   
-  raid = read_file("| /sbin/raiddetect -s 2>/dev/null", 0, 0);
+  raid = read_file("| /sbin/dmraid -rc 2>/dev/null", 0, 0);
 
   for(sl = raid; sl; sl = sl->next) {
-    s = hd_sysfs_id(sl->str);
+    s = *sl->str ? strchr(sl->str + 1, '/') : NULL;
     if(s) {
-      sl1 = add_str_list(&raid_sysfs, s);
+      sl1 = add_str_list(&raid_sysfs, NULL);
+      str_printf(&sl1->str, 0, "/block%s", s);
       len = strlen(sl1->str);
       if(len) sl1->str[len - 1] = 0;
     }
@@ -1003,5 +928,181 @@ void int_softraid(hd_data_t *hd_data)
   free_str_list(raid_sysfs);
 
 }
+
+
+/*
+ *
+ */
+void int_system(hd_data_t *hd_data)
+{
+  hd_t *hd_sys;
+  hd_smbios_t *sm;
+  struct {
+    unsigned notebook:1;
+    enum { v_none = 0, v_ibm = 1, v_toshiba, v_sony } vendor;
+  } is = { };
+  char *s;
+  struct stat sbuf;
+
+  for(hd_sys = hd_data->hd; hd_sys; hd_sys = hd_sys->next) {
+    if(
+      hd_sys->base_class.id == bc_internal &&
+      hd_sys->sub_class.id == sc_int_sys
+    ) break;
+  }
+
+  if(!hd_sys) return;
+
+  if(
+    hd_sys->vendor.name &&
+    !strncasecmp(hd_sys->vendor.name, "sony", sizeof "sony" - 1)
+  ) {
+    is.vendor = v_sony;
+  }
+
+  if(
+    hd_sys->device.name &&
+    !strncmp(hd_sys->device.name, "PCG-", sizeof "PCG-" - 1)
+  ) {
+    is.notebook = 1;
+  }
+
+  for(sm = hd_data->smbios; sm; sm = sm->next) {
+    if(
+      sm->any.type == sm_sysinfo &&
+      sm->sysinfo.manuf &&
+      !strcasecmp(sm->sysinfo.manuf, "ibm")
+    ) {
+      is.vendor = v_ibm;
+    }
+
+    if(
+      sm->any.type == sm_sysinfo &&
+      sm->sysinfo.manuf &&
+      !strcasecmp(sm->sysinfo.manuf, "toshiba")
+    ) {
+      is.vendor = v_toshiba;
+
+      if(!hd_sys->device.name && !hd_sys->device.id && sm->sysinfo.product) {
+        hd_sys->device.name = new_str(sm->sysinfo.product);
+      }
+      if(!hd_sys->vendor.name && !hd_sys->vendor.id) {
+        hd_sys->vendor.name = new_str("Toshiba");
+      }
+    }
+
+    if(
+      sm->any.type == sm_sysinfo &&
+      sm->sysinfo.manuf &&
+      !strncasecmp(sm->sysinfo.manuf, "sony", sizeof "sony" - 1)
+    ) {
+      is.vendor = v_sony;
+
+      if(!hd_sys->device.name && !hd_sys->device.id && sm->sysinfo.product) {
+        hd_sys->device.name = new_str(sm->sysinfo.product);
+        if(
+          (s = strchr(hd_sys->device.name, '(')) &&
+          hd_sys->device.name[strlen(hd_sys->device.name) - 1] == ')'
+        ) {
+          *s = 0;
+        }
+      }
+      if(!hd_sys->vendor.name && !hd_sys->vendor.id) {
+        hd_sys->vendor.name = new_str("Sony");
+      }
+    }
+
+    if(
+      sm->any.type == sm_chassis &&
+      (
+        (sm->chassis.ch_type.id >= 8 && sm->chassis.ch_type.id <= 11) ||
+        sm->chassis.ch_type.id == 14
+      )
+    ) {
+      is.notebook = 1;
+    }
+  }
+
+  ADD2LOG(
+    "  system type:%s%s\n",
+    is.vendor == v_ibm ? " ibm" :
+      is.vendor == v_toshiba ? " toshiba" :
+      is.vendor == v_sony ? " sony" :
+      "",
+    is.notebook ? " notebook" : ""
+  );
+
+  if(is.notebook && is.vendor) {
+    hd_sys->compat_vendor.id = MAKE_ID(TAG_SPECIAL, 0xf001);
+    hd_sys->compat_device.id = MAKE_ID(TAG_SPECIAL, is.vendor);
+  }
+
+  hd_sys->is.with_acpi = stat("/proc/acpi", &sbuf) ? 0 : 1;
+  ADD2LOG("  acpi: %d\n", hd_sys->is.with_acpi);
+}
+
+
+void int_legacy_geo(hd_data_t *hd_data)
+{
+  hd_t *hd;
+  hd_res_t *res;
+  int id;
+  char *s;
+  edd_info_t *ei;
+
+  if(!hd_data->edd) return;
+
+  for(hd = hd_data->hd; hd; hd = hd->next) {
+    if(
+      hd->base_class.id == bc_storage_device &&
+      hd->sub_class.id == sc_sdev_disk &&
+      hd->rom_id
+    ) {
+      id = strtol(hd->rom_id, &s, 0) - 0x80;
+      if(*s || id < 0 || id >= sizeof hd_data->edd / sizeof *hd_data->edd) continue;
+
+      ei = hd_data->edd + id;
+
+      if(ei->edd.cyls) {
+        res = add_res_entry(&hd->res, new_mem(sizeof *res));
+        res->disk_geo.type = res_disk_geo;
+        res->disk_geo.cyls = ei->edd.cyls;
+        res->disk_geo.heads = ei->edd.heads;
+        res->disk_geo.sectors = ei->edd.sectors;
+        res->disk_geo.size = ei->sectors;
+        res->disk_geo.geotype = geo_bios_edd;
+      }
+
+      if(ei->legacy.cyls) {
+        res = add_res_entry(&hd->res, new_mem(sizeof *res));
+        res->disk_geo.type = res_disk_geo;
+        res->disk_geo.cyls = ei->legacy.cyls;
+        res->disk_geo.heads = ei->legacy.heads;
+        res->disk_geo.sectors = ei->legacy.sectors;
+        res->disk_geo.geotype = geo_bios_legacy;
+      }
+
+    }
+  }
+}
 #endif
 
+
+void int_find_parent(hd_data_t *hd_data)
+{
+  hd_t *hd, *hd2;
+
+  for(hd = hd_data->hd; hd; hd = hd->next) {
+    if(hd->attached_to || !hd->parent_udi) continue;
+
+    for(hd2 = hd_data->hd; hd2; hd2 = hd2->next) {
+      if(!hd2->udi) continue;
+      if(!strcmp(hd->parent_udi, hd2->udi)) {
+        hd->attached_to = hd2->idx;
+        break;
+      }
+    }
+  }
+}
+
+
