@@ -137,7 +137,9 @@ cdb_isdn_card *get_isdn_info(hd_t *hd)
 
   if(hd->bus.id == bus_pci ||
     hd->bus.id == bus_isa ||
-    hd->bus.id == bus_usb) {
+    hd->bus.id == bus_usb ||
+    hd->bus.id == bus_pcmcia ||
+    hd->bus.id == bus_cardbus) {
 
     cic = NULL;
     u0 = ID_VALUE(hd->vendor.id);
@@ -185,6 +187,25 @@ cdb_isdn_card *get_isdn_info(hd_t *hd)
       	cic = hd_cdbisdn_get_card_from_id(ID_VALUE(hd->vendor.id), ID_VALUE(hd->device.id),
       				0xffff, 0xffff);
     }
+    
+    if((hd->bus.id == bus_pcmcia || hd->bus.id == bus_cardbus) &&
+    	(hd->base_class.id == bc_network || hd->base_class.id == bc_isdn)) {
+    	if (hd->drivers && hd->drivers->str) {
+    		if (0 == strcmp(hd->drivers->str, "teles_cs")) {
+    			cic = hd_cdbisdn_get_card_from_type(8, 0);
+    		} else if (0 == strcmp(hd->drivers->str, "sedlbauer_cs")) {
+    			cic = hd_cdbisdn_get_card_from_type(22, 2);
+    		} else if (0 == strcmp(hd->drivers->str, "avma1_cs")) {
+    			cic = hd_cdbisdn_get_card_from_type(26, 0);
+    		} else if (0 == strcmp(hd->drivers->str, "fcpcmcia_cs")) {
+    			cic = hd_cdbisdn_get_card_from_type(8002, 5);
+    		} else if (0 == strcmp(hd->drivers->str, "elsa_cs")) {
+    			cic = hd_cdbisdn_get_card_from_type(10, 11);
+    		} else if (0 == strcmp(hd->drivers->str, "avm_cs")) {
+    			cic = hd_cdbisdn_get_card_from_type(8001, 2);
+    		}
+        }
+    }
 
     if (cic && cic->Class && strcmp(cic->Class, "DSL")) {
       cic0 = new_mem(sizeof *cic0);
@@ -270,9 +291,9 @@ cdb_isdn_card *get_dsl_info(hd_t *hd)
       hd->sub_class.id = sc_dsl_unknown;
       civ = hd_cdbisdn_get_vario(cic->vario);
       if (civ && civ->interface) {
-        if (!strcmp(civ->interface, "CAPI20")) {
+        if (0 == strncmp(civ->interface, "CAPI20", 6)) {
           hd->sub_class.id = sc_dsl_capi;
-        } else if (!strcmp(civ->interface, "pppoe")) {
+        } else if (0 == strncmp(civ->interface, "pppoe", 5)) {
           hd->sub_class.id = sc_dsl_pppoe;
         }
       }
