@@ -141,7 +141,7 @@ void hd_scan_misc(hd_data_t *hd_data)
               res->disk_geo.cyls = geo.cylinders;
               res->disk_geo.heads = geo.heads;
               res->disk_geo.sectors = geo.sectors;
-              res->disk_geo.logical = 1;
+              res->disk_geo.geotype = geo_logical;
               size = geo.cylinders * geo.heads * geo.sectors;
               for(res = hd->res; res; res = res->next) {
                 if(res->any.type == res_size && res->size.unit == size_unit_sectors) {
@@ -345,6 +345,8 @@ void hd_scan_misc2(hd_data_t *hd_data)
       hd = add_hd_entry(hd_data, __LINE__, 0);
       hd->base_class.id = bc_storage;
       hd->sub_class.id = sc_sto_ide;
+      hd->compat_vendor.id = MAKE_ID(TAG_SPECIAL, 0x3000);
+      hd->compat_device.id = MAKE_ID(TAG_SPECIAL, 0x1000);
       /* use join_res to join the i/o ranges of ide0/1 */
       join_res_io(&hd->res, res);
       join_res_irq(&hd->res, res);
@@ -569,6 +571,7 @@ void read_irqs(misc_t *m)
 
   for(sl = m->proc_irq; sl; sl = sl->next) {
     /* irq */
+    i = 0;
     if(sscanf(sl->str, " %u: %n", &u, &i) < 1) continue;
     v = 0;
     j = i;
