@@ -33,9 +33,11 @@
 
 static void get_usb_devs(hd_data_t *hd_data);
 static void set_class_entries(hd_data_t *hd_data, hd_t *hd, usb_t *usb);
+static void add_input_dev(hd_data_t *hd_data, char *name);
 static void get_input_devs(hd_data_t *hd_data);
 static void get_printer_devs(hd_data_t *hd_data);
 static void read_usb_lp(hd_data_t *hd_data, hd_t *hd);
+static void get_serial_devs(hd_data_t *hd_data);
 
 void hd_scan_sysfs_usb(hd_data_t *hd_data)
 {
@@ -68,12 +70,15 @@ void hd_scan_sysfs_usb(hd_data_t *hd_data)
   PROGRESS(3, 4, "lp");
   get_printer_devs(hd_data);
 
+  PROGRESS(3, 5, "serial");
+  get_serial_devs(hd_data);
+
 }
 
 
 void get_usb_devs(hd_data_t *hd_data)
 {
-  uint64_t ul0, ul1;
+  uint64_t ul0;
   unsigned u1, u2, u3;
   hd_t *hd, *hd1;
   usb_t *usb;
@@ -113,9 +118,7 @@ void get_usb_devs(hd_data_t *hd_data)
     );
 
     if(
-      hd_attr_uint(sysfs_get_device_attr(sf_dev, "bInterfaceNumber"), &ul0, 16) &&
-      hd_attr_uint(sysfs_get_device_attr(sf_dev, "bAlternateSetting"), &ul1, 0) &&
-      ul1 == 0
+      hd_attr_uint(sysfs_get_device_attr(sf_dev, "bInterfaceNumber"), &ul0, 16)
     ) {
       hd = add_hd_entry(hd_data, __LINE__, 0);
 
@@ -131,6 +134,16 @@ void get_usb_devs(hd_data_t *hd_data)
 
       usb->ifdescr = ul0;
 
+      if((s = hd_attr_str(sysfs_get_device_attr(sf_dev, "modalias")))) {
+        s = canon_str(s, strlen(s));
+        ADD2LOG("    modalias = \"%s\"\n", s);
+        if(s && *s) {
+          hd->modalias = s;
+          s = NULL;
+        }
+        s = free_mem(s);
+      }
+
       ADD2LOG("    bInterfaceNumber = %u\n", hd->func);
 
       if(hd_attr_uint(sysfs_get_device_attr(sf_dev, "bInterfaceClass"), &ul0, 16)) {
@@ -211,9 +224,9 @@ void get_usb_devs(hd_data_t *hd_data)
 
           if((s = hd_attr_str(sysfs_get_device_attr(sf_dev_2, "speed")))) {
             s = canon_str(s, strlen(s));
-            if(strcmp(s, "1.5")) usb->speed = 15*100000;
-            else if(strcmp(s, "12")) usb->speed = 12*1000000;
-            else if(strcmp(s, "240")) usb->speed = 240*1000000;
+            if(!strcmp(s, "1.5")) usb->speed = 15*100000;
+            else if(!strcmp(s, "12")) usb->speed = 12*1000000;
+            else if(!strcmp(s, "480")) usb->speed = 480*1000000;
             ADD2LOG("    speed = \"%s\"\n", s);
             s = free_mem(s);
           }
@@ -362,7 +375,13 @@ void set_class_entries(hd_data_t *hd_data, hd_t *hd, usb_t *usb)
 
   switch(cls) {
     case 2:
-      hd->base_class.id = bc_modem;
+      if(usb->i_sub == 6 && usb->i_prot == 0) {
+        hd->base_class.id = bc_network;
+        hd->sub_class.id = 0x91;
+      }
+      else if(usb->i_sub == 2 && usb->i_prot >= 1 && usb->i_prot <= 6) {
+        hd->base_class.id = bc_modem;
+      }
       break;
 
     case 3:
@@ -423,6 +442,13 @@ void set_class_entries(hd_data_t *hd_data, hd_t *hd, usb_t *usb)
       hd->base_class.id = bc_hub;
       break;
 
+    case 0xe0:
+      if(sub == 1 && prot == 1) {
+        hd->base_class.id = bc_bluetooth;
+        hd->sub_class.id = 0;
+      }
+      break;
+
     case 0xff:
       /* hp psc 2100, 2200, 2150, officejet 6100 */
       if(
@@ -448,87 +474,122 @@ void set_class_entries(hd_data_t *hd_data, hd_t *hd, usb_t *usb)
     hd->sub_class.id = u & 0xff;
   }
 
+  /* FIXME: hack for bt isdn box */
+  if(
+    hd->vendor.id == MAKE_ID(TAG_USB, 0x057c) &&
+    hd->device.id == MAKE_ID(TAG_USB, 0x2200)
+  ) {
+    hd_set_hw_class(hd, hw_bluetooth);
+  }
+
 }
 
 
-void get_input_devs(hd_data_t *hd_data)
+void add_input_dev(hd_data_t *hd_data, char *name)
 {
   hd_t *hd;
   char *s, *t;
-  hd_dev_num_t dev_num;
+  hd_dev_num_t dev_num = { };
   unsigned u1, u2;
 
-  struct sysfs_class *sf_class;
   struct sysfs_class_device *sf_cdev;
   struct sysfs_device *sf_dev;
-  struct dlist *sf_cdev_list;
 
-  sf_class = sysfs_open_class("input");
+  if(!(sf_cdev = sysfs_open_class_device_path(name))) return;
 
-  if(!sf_class) {
-    ADD2LOG("sysfs: no such class: input\n");
+  ADD2LOG(
+    "  input: name = %s, path = %s\n",
+    sf_cdev->name,
+    hd_sysfs_id(sf_cdev->path)
+  );
+
+  if(!strncmp(sf_cdev->name, "ts", sizeof "ts" - 1)) {
+    sysfs_close_class_device(sf_cdev);
     return;
   }
 
-  sf_cdev_list = sysfs_get_class_devices(sf_class);
-  if(sf_cdev_list) dlist_for_each_data(sf_cdev_list, sf_cdev, struct sysfs_class_device) {
+  if((s = hd_attr_str(sysfs_get_classdev_attr(sf_cdev, "dev")))) {
+    if(sscanf(s, "%u:%u", &u1, &u2) == 2) {
+      dev_num.type = 'c';
+      dev_num.major = u1;
+      dev_num.minor = u2;
+      dev_num.range = 1;
+    }
+    ADD2LOG("    dev = %u:%u\n", u1, u2);
+  }
+
+  sf_dev = sysfs_get_classdev_device(sf_cdev);
+  if(sf_dev) {
+    s = hd_sysfs_id(sf_dev->path);
+
     ADD2LOG(
-      "  input: name = %s, path = %s\n",
-      sf_cdev->name,
-      hd_sysfs_id(sf_cdev->path)
+      "    input device: bus = %s, bus_id = %s driver = %s\n      path = %s\n",
+      sf_dev->bus,
+      sf_dev->bus_id,
+      sf_dev->driver_name,
+      s
     );
 
-    if((s = hd_attr_str(sysfs_get_classdev_attr(sf_cdev, "dev")))) {
-      if(sscanf(s, "%u:%u", &u1, &u2) == 2) {
-        dev_num.type = 'c';
-        dev_num.major = u1;
-        dev_num.minor = u2;
-        dev_num.range = 1;
-      }
-      ADD2LOG("    dev = %u:%u\n", u1, u2);
-    }
+    for(hd = hd_data->hd; hd; hd = hd->next) {
+      if(
+        hd->module == hd_data->module &&
+        hd->sysfs_id &&
+        s &&
+        !strcmp(s, hd->sysfs_id)
+      ) {
+        t = NULL;
+        str_printf(&t, 0, "/dev/input/%s", sf_cdev->name);
 
-    sf_dev = sysfs_get_classdev_device(sf_cdev);
-    if(sf_dev) {
-      s = hd_sysfs_id(sf_dev->path);
+        if(strncmp(sf_cdev->name, "mouse", sizeof "mouse" - 1)) {
+          hd->unix_dev_name = t;
+          hd->unix_dev_num = dev_num;
+        }
+        else {
+          hd->unix_dev_name2 = t;
+          hd->unix_dev_num2 = dev_num;
 
-      ADD2LOG(
-        "    input device: bus = %s, bus_id = %s driver = %s\n      path = %s\n",
-        sf_dev->bus,
-        sf_dev->bus_id,
-        sf_dev->driver_name,
-        s
-      );
-
-      for(hd = hd_data->hd; hd; hd = hd->next) {
-        if(
-          hd->module == hd_data->module &&
-          hd->sysfs_id &&
-          s &&
-          !strcmp(s, hd->sysfs_id)
-        ) {
-          t = NULL;
-          str_printf(&t, 0, "/dev/input/%s", sf_cdev->name);
-
-          if(strncmp(sf_cdev->name, "mouse", sizeof "mouse" - 1)) {
-            hd->unix_dev_name = t;
-            hd->unix_dev_num = dev_num;
-          }
-          else {
-            hd->unix_dev_name2 = t;
-            hd->unix_dev_num2 = dev_num;
-
-            dev_num.major = 13;
-            dev_num.minor = 63;
-            hd->unix_dev_name = new_str(DEV_MICE);
-            hd->unix_dev_num = dev_num;
-          }
+          dev_num.major = 13;
+          dev_num.minor = 63;
+          hd->unix_dev_name = new_str(DEV_MICE);
+          hd->unix_dev_num = dev_num;
         }
       }
     }
   }
 
-  sysfs_close_class(sf_class);
+  sysfs_close_class_device(sf_cdev);
+}
+
+
+void get_input_devs(hd_data_t *hd_data)
+{
+  struct sysfs_link *sf_link;
+  struct sysfs_directory *sf_dir, *sf_subdir;
+  struct dlist *sf_list;
+
+  sf_dir = sysfs_open_directory("/sys/class/input");
+
+  if(!sf_dir) {
+    ADD2LOG("sysfs: no such class: input\n");
+    return;
+  }
+
+  /*
+   * A bit tricky: if there are links, assume newer sysfs layout with compat
+   * symlinks; if not, assume old layout with directories.
+   */
+
+  sysfs_read_dir_links(sf_dir);
+  if(!sf_dir->links) sysfs_read_dir_subdirs(sf_dir);
+
+  if((sf_list = sf_dir->links)) dlist_for_each_data(sf_list, sf_link, struct sysfs_link) {
+    add_input_dev(hd_data, sf_link->target);
+  }
+  else if((sf_list = sf_dir->subdirs)) dlist_for_each_data(sf_list, sf_subdir, struct sysfs_directory) {
+    add_input_dev(hd_data, sf_subdir->path);
+  }
+
+  sysfs_close_directory(sf_dir);
 }
 
 
@@ -536,7 +597,7 @@ void get_printer_devs(hd_data_t *hd_data)
 {
   hd_t *hd;
   char *s, *t;
-  hd_dev_num_t dev_num;
+  hd_dev_num_t dev_num = { };
   unsigned u1, u2;
 
   struct sysfs_class *sf_class;
@@ -675,3 +736,83 @@ void read_usb_lp(hd_data_t *hd_data, hd_t *hd)
   close(fd);
 }
 #undef MATCH_FIELD
+
+
+void get_serial_devs(hd_data_t *hd_data)
+{
+  hd_t *hd;
+  char *s, *t;
+  hd_dev_num_t dev_num = { };
+  unsigned u1, u2;
+
+  struct sysfs_class *sf_class;
+  struct sysfs_class_device *sf_cdev;
+  struct sysfs_device *sf_dev;
+  struct dlist *sf_cdev_list;
+
+  sf_class = sysfs_open_class("tty");
+
+  if(!sf_class) {
+    ADD2LOG("sysfs: no such class: tty\n");
+    return;
+  }
+
+  sf_cdev_list = sysfs_get_class_devices(sf_class);
+  if(sf_cdev_list) dlist_for_each_data(sf_cdev_list, sf_cdev, struct sysfs_class_device) {
+    if(strncmp(sf_cdev->name, "ttyUSB", 6)) continue;
+
+    ADD2LOG(
+      "  usb: name = %s, path = %s\n",
+      sf_cdev->name,
+      hd_sysfs_id(sf_cdev->path)
+    );
+
+    if((s = hd_attr_str(sysfs_get_classdev_attr(sf_cdev, "dev")))) {
+      if(sscanf(s, "%u:%u", &u1, &u2) == 2) {
+        dev_num.type = 'c';
+        dev_num.major = u1;
+        dev_num.minor = u2;
+        dev_num.range = 1;
+      }
+      ADD2LOG("    dev = %u:%u\n", u1, u2);
+    }
+
+    sf_dev = sysfs_get_classdev_device(sf_cdev);
+    if(sf_dev) {
+      s = hd_sysfs_id(sf_dev->path);
+
+      if((t = strrchr(s, '/')) && !strncmp(t + 1, "ttyUSB", sizeof "ttyUSB" - 1)) *t =0;
+
+      ADD2LOG(
+        "    usb device: bus = %s, bus_id = %s driver = %s\n      path = %s\n",
+        sf_dev->bus,
+        sf_dev->bus_id,
+        sf_dev->driver_name,
+        s
+      );
+
+      for(hd = hd_data->hd; hd; hd = hd->next) {
+        if(
+          hd->module == hd_data->module &&
+          hd->sysfs_id &&
+          s &&
+          !strcmp(s, hd->sysfs_id)
+        ) {
+          t = NULL;
+          str_printf(&t, 0, "/dev/%s", sf_cdev->name);
+
+          hd->unix_dev_name = t;
+          hd->unix_dev_num = dev_num;
+
+          hd->base_class.id = bc_comm;
+          hd->sub_class.id = sc_com_ser;
+          hd->prog_if.id = 0x80;
+        }
+      }
+    }
+  }
+
+  sysfs_close_class(sf_class);
+}
+
+
